# SPDX-License-Identifier: Apache-2.0
import logging

import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.vm as utils_vm
from saltext.vmware.modules.tag import update

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_tag"
__proxyenabled__ = ["vmware_tag"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def present(name, description=None, category_id=None):
    """
    Create or update a tag instance.

    name
        Name of tag.

    description
        Description of tag.

    category_id
        (string) Category ID of type: com.vmware.cis.tagging.Tag. Only optional when updating
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    res = connect.request(
        "/rest/com/vmware/cis/tagging/tag", "GET", opts=__opts__, pillar=__pillar__
    )
    response = res["response"].json()
    token = res["token"]
    found = None
    for tag in response["value"]:
        url = f"/rest/com/vmware/cis/tagging/tag/id:{tag}"
        tag_ref = connect.request(url, "GET", token=token, opts=__opts__, pillar=__pillar__)
        tag_ref = tag_ref["response"].json()
        if tag_ref["value"]["name"] == name:
            found = tag_ref["value"]
            break
    if found:
        if description == found["description"]:
            ret["comment"] = "tag exists"
            return ret
        else:
            ret["changes"]["new"] = {}
            ret["changes"]["old"] = {}
            ret["changes"]["old"]["description"] = found["description"]
            ret["changes"]["new"]["description"] = description
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = f"{name} tag will be updated"
                return ret
            id = found["id"]
            spec = {"update_spec": {"description": description}}
            url = f"/rest/com/vmware/cis/tagging/tag/id:{id}"
            updated = connect.request(
                url, "PATCH", body=spec, token=token, opts=__opts__, pillar=__pillar__
            )
            if updated["response"].status_code == 200:
                ret["comment"] = "updated"
                return ret
            ret["status_code"] = updated["response"].status_code
            ret["reason"] = updated["response"].reason
            ret["comment"] = "failed to update"
            ret["result"] = False
            return ret
    else:
        if category_id:
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = f"{name} tag will be created"
                return ret
            data = {
                "create_spec": {
                    "category_id": category_id,
                    "description": description,
                    "name": name,
                }
            }
            create = connect.request(
                "/rest/com/vmware/cis/tagging/tag",
                "POST",
                body=data,
                token=token,
                opts=__opts__,
                pillar=__pillar__,
            )
            response = create["response"].json()
            ret["changes"]["tag_id"] = response["value"]
            ret["comment"] = "created"
            return ret
        else:
            ret["result"] = False
            ret["comment"] = "category_id required to create a tag"
            return ret


def absent(name):
    """
    Delete tag.

    name
        Name of tag.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    res = connect.request(
        "/rest/com/vmware/cis/tagging/tag", "GET", opts=__opts__, pillar=__pillar__
    )
    response = res["response"].json()
    token = res["token"]
    found = None
    for tag in response["value"]:
        url = f"/rest/com/vmware/cis/tagging/tag/id:{tag}"
        tag_ref = connect.request(url, "GET", token=token, opts=__opts__, pillar=__pillar__)
        tag_ref = tag_ref["response"].json()
        if tag_ref["value"]["name"] == name:
            found = tag_ref["value"]
            break
    if found:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"{name} tag will be deleted"
            return ret
        id = found["id"]
        url = f"/rest/com/vmware/cis/tagging/tag/id:{id}"
        delete = connect.request(url, "DELETE", token=token, opts=__opts__, pillar=__pillar__)
        if delete["response"].status_code == 200:
            ret["comment"] = "deleted"
            return ret
        ret["status_code"] = delete["response"].status_code
        ret["reason"] = delete["response"].reason
        ret["comment"] = "failed to delete"
        ret["result"] = False
        return ret
    else:
        ret["comment"] = "Tag does not exist"
        return ret


def present_category(name, associable_types, cardinality, description=""):
    """
    Create or update a category.

    name
        The display name of the category.

    associable_types
        (list) Object types to which this category’s tags can be attached.

    cardinality
        The CategoryModel.Cardinality enumerated type defines the number of tags in a category that can be assigned to an object. SINGLE, MULTIPLE

    description
        (optional) The description of the category.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    res = connect.request(
        "/rest/com/vmware/cis/tagging/category", "GET", opts=__opts__, pillar=__pillar__
    )
    response = res["response"].json()
    token = res["token"]
    found = None
    for cat in response["value"]:
        url = f"/rest/com/vmware/cis/tagging/category/id:{cat}"
        cat_ref = connect.request(url, "GET", token=token, opts=__opts__, pillar=__pillar__)
        cat_ref = cat_ref["response"].json()
        if cat_ref["value"]["name"] == name:
            found = cat_ref["value"]
            break
    if found:
        if (
            associable_types == found["associable_types"]
            and cardinality == found["cardinality"]
            and description == found["description"]
        ):
            ret["comment"] = "category exists"
            return ret
        else:
            ret["changes"]["new"] = {}
            ret["changes"]["old"] = {}
            if associable_types != found["associable_types"]:
                ret["changes"]["old"]["associable_types"] = found["associable_types"]
                ret["changes"]["new"]["associable_types"] = associable_types
            if cardinality != found["cardinality"]:
                ret["changes"]["old"]["cardinality"] = found["cardinality"]
                ret["changes"]["new"]["cardinality"] = cardinality
            if description != found["description"]:
                ret["changes"]["old"]["description"] = found["description"]
                ret["changes"]["new"]["description"] = description
            if __opts__["test"]:
                ret["result"] = None
                ret["comment"] = f"{name} category will be updated"
                return ret
            id = found["id"]
            spec = {"update_spec": {}}
            if associable_types:
                spec["update_spec"]["associable_types"] = associable_types
            if cardinality:
                spec["update_spec"]["cardinality"] = cardinality
            if description:
                spec["update_spec"]["description"] = description
            url = f"/rest/com/vmware/cis/tagging/category/id:{id}"
            updated = connect.request(url, "PATCH", body=spec, opts=__opts__, pillar=__pillar__)
            if updated["response"].status_code == 200:
                ret["comment"] = "updated"
                return ret
            ret["status_code"] = updated["response"].status_code
            ret["reason"] = updated["response"].reason
            ret["comment"] = "failed to update"
            ret["result"] = False
            return ret
    else:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"{name} category will be created"
            return ret
        data = {
            "create_spec": {
                "associable_types": associable_types,
                "cardinality": cardinality,
                "description": description,
                "name": name,
            }
        }
        create = connect.request(
            "/rest/com/vmware/cis/tagging/category",
            "POST",
            body=data,
            opts=__opts__,
            pillar=__pillar__,
        )
        response = create["response"].json()
        ret["changes"]["category_id"] = response["value"]
        ret["comment"] = "created"
        return ret


def absent_category(name):
    """
    Delete category.

    name
        Name of category.
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    res = connect.request(
        "/rest/com/vmware/cis/tagging/category", "GET", opts=__opts__, pillar=__pillar__
    )
    response = res["response"].json()
    token = res["token"]
    found = None
    for cat in response["value"]:
        url = f"/rest/com/vmware/cis/tagging/category/id:{cat}"
        cat_ref = connect.request(url, "GET", token=token, opts=__opts__, pillar=__pillar__)
        cat_ref = cat_ref["response"].json()
        if cat_ref["value"]["name"] == name:
            found = cat_ref["value"]
            break
    if found:
        if __opts__["test"]:
            ret["result"] = None
            ret["comment"] = f"{name} category will be deleted"
            return ret
        id = found["id"]
        url = f"/rest/com/vmware/cis/tagging/category/id:{id}"
        delete = connect.request(url, "DELETE", opts=__opts__, pillar=__pillar__)
        if delete["response"].status_code == 200:
            ret["comment"] = "deleted"
            return ret
        ret["status_code"] = delete["response"].status_code
        ret["reason"] = delete["response"].reason
        ret["comment"] = "failed to delete"
        ret["result"] = False
        return ret
    else:
        ret["comment"] = "Category does not exist"
        return ret



"""
Salt execution module for ESX hosts
Provides methods to Add/Remove one or more ESX hosts in the target SDDC

"""
import logging

from saltext.vmware.modules import vmc_sddc
from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates

log = logging.getLogger(__name__)

__virtualname__ = "vmc_sddc_host"
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def manage(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    num_hosts,
    availability_zone=None,
    cluster_id=None,
    esxs=None,
    strict_placement=False,
    action=None,
    verify_ssl=True,
    cert=None,
):
    """
    Add/remove host for a given SDDC

    Please refer the `VMC Create ESXs documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/esxs/post/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

        salt <minion-key-id> vmc_sddc_host.manage hostname=vmc.vmware.com ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the host would be added/removed

    num_hosts: Integer
        (Required) Count of Hosts that would be added/removed for the given SDDC

    availability_zone: String
        (Optional) Availability zone where the hosts should be provisioned.
        (Can be specified only for privileged host operations).

    cluster_id: String
        (Optional) An optional cluster id if the esxs operation has to be on a specific cluster.

    esxs: Array Of String As UUID
        (Optional) An optional list of ESX IDs to remove.

    strict_placement: Boolean
        (Optional) An option to indicate if the host needs to be strictly placed in a placement group.
        Fail the operation otherwise.

    action: String
        (Optional)
        If action = 'add', will add the esx.

        If action = 'remove', will delete the esx/esxs bound to a single cluster(Cluster Id is mandatory for non cluster 1 esx remove).

        If action = 'force-remove', will delete the esx even if it can lead to data loss (This is an privileged operation).

        If action = 'addToAll', will add esxs to all clusters in the SDDC (This is an privileged operation).

        If action = 'removeFromAll', will delete the esxs from all clusters in the SDDC (This is an privileged operation).

        If action = 'attach-diskgroup', will attach the provided diskgroups to a given host (privileged).

        If action = 'detach-diskgroup', will detach the diskgroups of a given host (privileged).

        Default behaviour is 'add'

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    For example:
        .. code::

            {
                "availability_zone": "us-west-2a",
                "cluster_id": "e97920ae-1410-4269-9caa-29584eb8cf6d",
                "esxs": [
                "94ce40e1-8619-45b5-9817-0f3466d0dc78"
                ],
                "num_hosts": 1,
                "strict_placement": false
            }

    """

    log.info("Managing host for the SDDC %s", sddc_id)
    api_base_url = vmc_request.set_base_url(hostname)
    api_url = "{base_url}vmc/api/orgs/{org_id}/sddcs/{sddc_id}/esxs".format(
        base_url=api_base_url, org_id=org_id, sddc_id=sddc_id
    )

    allowed_dict = {
        "num_hosts": num_hosts,
        "availability_zone": availability_zone,
        "cluster_id": cluster_id,
        "esxs": esxs,
        "strict_placement": strict_placement,
    }
    req_data = vmc_request._filter_kwargs(allowed_kwargs=allowed_dict.keys(), **allowed_dict)
    params = vmc_request._filter_kwargs(allowed_kwargs=["action"], action=action)
    request_data = vmc_request.create_payload_for_request(vmc_templates.manage_sddc_host, req_data)
    return vmc_request.call_api(
        method=vmc_constants.POST_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_sddc_host.manage",
        data=request_data,
        params=params,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def list_(hostname, refresh_key, authorization_host, org_id, sddc_id, verify_ssl=True, cert=None):
    """
    Retrieves ESX hosts list for the given SDDC

    Please refer the `VMC Get SDDC documentation <https://developer.vmware.com/docs/vmc/latest/vmc/api/orgs/org/sddcs/sddc/get/>`_ to get insight of functionality and input parameters

    CLI Example:

    .. code-block:: bash

    salt <minion-key-id> vmc_sddc_host.get hostname=vmc.vmware.com ...

    hostname
        The host name of VMC

    refresh_key
        API Token of the user which is used to get the Access Token required for VMC operations

    authorization_host
        Hostname of the Cloud Services Platform (CSP)

    org_id
        The Id of organization to which the SDDC belongs to

    sddc_id
        The Id of SDDC for which the hosts would be retrieved

    verify_ssl
    (Optional) Option to enable/disable SSL verification. Enabled by default.
    If set to False, the certificate validation is skipped.

    cert
    (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
    The certificate can be retrieved from browser.

    """

    log.info("Retrieving hosts for SDDC {} %s", sddc_id)
    sddc_detail = vmc_sddc.get_by_id(
        hostname=hostname,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        org_id=org_id,
        sddc_id=sddc_id,
        verify_ssl=verify_ssl,
        cert=cert,
    )
    if "error" in sddc_detail:
        return sddc_detail
    cluster_list = sddc_detail["resource_config"]["clusters"]
    esx_hosts_details = []
    for cluster in cluster_list:
        esx_hosts_details += cluster["esx_host_list"]
    result = {"description": "vmc_sddc_host.list_", "esx_hosts": esx_hosts_details}
    return result



"""
Salt execution module for security groups
Provides methods to Create, Read, Update and Delete security groups.
"""
import logging
import os

from saltext.vmware.utils import vmc_constants
from saltext.vmware.utils import vmc_request
from saltext.vmware.utils import vmc_templates

log = logging.getLogger(__name__)

__virtualname__ = "vmc_security_groups"


def __virtual__():
    return __virtualname__


def _create_payload_for_security_group(security_group_id, domain_id, user_input):
    """
    This function creates the payload based on the template and user input passed
    """
    template_data = getattr(
        vmc_templates,
        "create_security_groups_" + domain_id,
        vmc_templates.create_security_groups_cgw,
    )
    data = vmc_request.create_payload_for_request(template_data, user_input)
    data["id"] = data["display_name"] = security_group_id
    return data


def get(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    verify_ssl=True,
    cert=None,
    cursor=None,
    page_size=None,
    sort_by=None,
    sort_ascending=None,
):
    """
    Retrieves security groups from Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_groups.get hostname=nsxt-manager.local  ...


    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id from which security groupss should be retrieved

    domain_id
        The domain_id for which the security groups should be retrieved

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)

    page_size
        (Optional) Maximum number of results to return in this page. Default page size is 1000.

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order. Enabled by default.

    """
    log.info("Retrieving security groups for SDDC %s", sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/groups"
    )
    api_url = api_url.format(
        base_url=api_url_base, org_id=org_id, sddc_id=sddc_id, domain_id=domain_id
    )

    params = vmc_request._filter_kwargs(
        allowed_kwargs=["cursor", "page_size", "sort_ascending", "sort_by"],
        cursor=cursor,
        page_size=page_size,
        sort_by=sort_by,
        sort_ascending=sort_ascending,
    )
    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_groups.get",
        verify_ssl=verify_ssl,
        cert=cert,
        params=params,
    )


def get_by_id(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_group_id,
    verify_ssl=True,
    cert=None,
):
    """
    Retrieves security groups from Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_groups.get_by_id hostname=nsxt-manager.local security_group_id ...

    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id from which security groups should be retrieved

    domain_id
        The domain_id for which the security groups should be retrieved

    scurity_group_id
        Id of the security group to be retrieved from SDDC

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.
    """
    log.info("Retrieving security group %s for SDDC %s", security_group_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/groups/{security_group_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_group_id=security_group_id,
    )

    return vmc_request.call_api(
        method=vmc_constants.GET_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_groups.get_by_id",
        verify_ssl=verify_ssl,
        cert=cert,
    )


def delete(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_group_id,
    verify_ssl=True,
    cert=None,
):
    """
    Delete security groups from Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_groups.delete hostname=nsxt-manager.local security_group_id=security_group_id ...

    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id from which security groups will be deleted

    domain_id
        The domain_id for which the security groups should be deleted

    security_group_id
        sepcific security groups id

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to VMC Cloud Console.
        The certificate can be retrieved from browser.

    """
    log.info("Deleting security group %s for SDDC %s", security_group_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/groups/{security_group_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_group_id=security_group_id,
    )
    return vmc_request.call_api(
        method=vmc_constants.DELETE_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_groups.delete",
        responsebody_applicable=False,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def create(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_group_id,
    verify_ssl=True,
    cert=None,
    expression=None,
    description=None,
    tags=vmc_constants.VMC_NONE,
):
    """
    Create security groups for Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_groups.create hostname=nsxt-manager.local public_ip_name=vmc_security_groups ...

    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id for which security groups should be retrieved.

    domain_id
        The domain_id for which the security groups should be retrieved.  Possible values: mgw and cgw

    security_group_id
        name of security groups it will create id same as name.

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    expression
        (Optional) Expression for security group members
            The expression list must follow below criteria:
            1. A non-empty expression list, must be of odd size. In a list, with
            indices starting from 0, all non-conjunction expressions must be at
            even indices, separated by a conjunction expression at odd
            indices.
            2. The total of ConditionExpression and NestedExpression in a list
            should not exceed 5.
            3. The total of IPAddressExpression, MACAddressExpression, external
            IDs in an ExternalIDExpression and paths in a PathExpression must not exceed
            500.
            4. Each expression must be a valid Expression. See the definition of
            the Expression type for more information.

        Its list of dicts

            Example values

            1. [{"member_type":"VirtualMachine","resource_type":"ExternalIDExpression",
            "external_ids":["52bf8bd0-95b1-2e58-5180-ccfa743da576"]}]

            2. [{"value":"Linux","member_type":"VirtualMachine","key":"OSName",
               "operator":"EQUALS","resource_type":"Condition"},
               {"resource_type":"ConjunctionOperator","conjunction_operator":"OR"},
               {"member_type":"VirtualMachine","resource_type":"ExternalIDExpression",
               "external_ids":["52bf8bd0-95b1-2e58-5180-ccfa743da576"]}]

            3. [{"ip_addresses" : ["203.0.113.1", "203.0.113.2"],
                  "resource_type" : "IPAddressExpression"} ]

            default value is []

    description
        (Optional) Description of Security Groups
            default value is ""

    tags
        (Optional) Opaque identifiers meaningful to the API user. Maximum 30 tags can be associated:
            default value is []

        .. code-block::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    Example values:

        .. code-block::

            {
                "expression": [
                  {
                    "member_type": "VirtualMachine",
                    "value": "webvm",
                    "key": "Tag",
                    "operator": "EQUALS",
                    "resource_type": "Condition"
                  }
                ],
                "description": "web group"
            }

    Please refer the `Security Groups <https://developer.vmware.com/docs/nsx-vmc-policy/latest/policy/api/v1/infra/domains/domain-id/groups/group-id/put/>`_ to get insight of input parameters

    """
    log.info("Creating security group %s for SDDC %s", security_group_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/groups/{security_group_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_group_id=security_group_id,
    )

    allowed_dict = {
        "expression": expression,
        "description": description,
        "tags": tags,
    }
    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = _create_payload_for_security_group(security_group_id, domain_id, req_data)
    return vmc_request.call_api(
        method=vmc_constants.PUT_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_groups.create",
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )


def update(
    hostname,
    refresh_key,
    authorization_host,
    org_id,
    sddc_id,
    domain_id,
    security_group_id,
    verify_ssl=True,
    cert=None,
    expression=None,
    description=None,
    tags=vmc_constants.VMC_NONE,
    display_name=None,
):
    """
    Update security groups for Given SDDC

    CLI Example:

    .. code-block:: bash

        salt vm_minion vmc_security_groups.update hostname=nsxt-manager.local public_ip_name=vmc_security_groups ...

    hostname
        The host name of NSX-T manager

    refresh_key
        refresh_key to get access token

    authorization_host
        hostname to get access token

    org_id
        org_id of the SDDC

    sddc_id
        sddc_id for which security groups should be retrieved

    domain_id
        The domain_id for which the security groups should be retrieved. Possible values: mgw and cgw

    security_group_id
        name of security groups it will update

    verify_ssl
        (Optional) Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    expression
        (Optional) Expression for security group members
            The expression list must follow below criteria:
            1. A non-empty expression list, must be of odd size. In a list, with
            indices starting from 0, all non-conjunction expressions must be at
            even indices, separated by a conjunction expression at odd
            indices.
            2. The total of ConditionExpression and NestedExpression in a list
            should not exceed 5.
            3. The total of IPAddressExpression, MACAddressExpression, external
            IDs in an ExternalIDExpression and paths in a PathExpression must not exceed
            500.
            4. Each expression must be a valid Expression. See the definition of
            the Expression type for more information.

        Its list of dicts
            Example values

            1. [{"member_type":"VirtualMachine","resource_type":"ExternalIDExpression",
            "external_ids":["52bf8bd0-95b1-2e58-5180-ccfa743da576"]}]

            2. [{"value":"Linux","member_type":"VirtualMachine","key":"OSName",
               "operator":"EQUALS","resource_type":"Condition"},
               {"resource_type":"ConjunctionOperator","conjunction_operator":"OR"},
               {"member_type":"VirtualMachine","resource_type":"ExternalIDExpression",
               "external_ids":["52bf8bd0-95b1-2e58-5180-ccfa743da576"]}]

            3. [{ip_addresses" : ["203.0.113.1", "203.0.113.2"],
                  "resource_type" : "IPAddressExpression"} ]

            default value is []

    description
        (Optional) Description of Security Groups
        default value is ""

    tags
        (Optional) Opaque identifiers meaningful to the API user. Maximum 30 tags can be associated:
        default value is [] empty list

        .. code-block::

            tags='[
                {
                    "tag": "<tag-key-1>"
                    "scope": "<tag-value-1>"
                },
                {
                    "tag": "<tag-key-2>"
                    "scope": "<tag-value-2>"
                }
            ]'

    display_name
        Identifier to use when displaying entity in logs or GUI

    Example values:

        .. code-block::

            {
                "expression": [
                  {
                    "member_type": "VirtualMachine",
                    "value": "webvm",
                    "key": "Tag",
                    "operator": "EQUALS",
                    "resource_type": "Condition"
                  }
                ],
                "description": "web group"
            }

    Please refer the `Security groups update <https://developer.vmware.com/docs/nsx-vmc-policy/latest/policy/api/v1/infra/domains/domain-id/groups/group-id/patch/>`_ to get insight of input parameters

    """
    log.info("Updating security group %s for SDDC %s", security_group_id, sddc_id)

    api_url_base = vmc_request.set_base_url(hostname)
    api_url = (
        "{base_url}vmc/reverse-proxy/api/orgs/{org_id}/sddcs/{sddc_id}/"
        "policy/api/v1/infra/domains/{domain_id}/groups/{security_group_id}"
    )
    api_url = api_url.format(
        base_url=api_url_base,
        org_id=org_id,
        sddc_id=sddc_id,
        domain_id=domain_id,
        security_group_id=security_group_id,
    )

    existing_data = get_by_id(
        hostname,
        refresh_key,
        authorization_host,
        org_id,
        sddc_id,
        domain_id,
        security_group_id,
        verify_ssl,
        cert,
    )

    if "error" in existing_data:
        return existing_data

    allowed_dict = {
        "display_name": display_name,
        "expression": expression,
        "description": description,
        "tags": tags,
    }
    req_data = vmc_request._filter_kwargs(
        allowed_kwargs=allowed_dict.keys(), allow_none=["tags"], **allowed_dict
    )

    payload = vmc_request.create_payload_for_request(
        vmc_templates.update_security_groups, req_data, existing_data
    )
    return vmc_request.call_api(
        method=vmc_constants.PATCH_REQUEST_METHOD,
        url=api_url,
        refresh_key=refresh_key,
        authorization_host=authorization_host,
        description="vmc_security_groups.update",
        responsebody_applicable=False,
        data=payload,
        verify_ssl=verify_ssl,
        cert=cert,
    )



# SPDX-License-Identifier: Apache-2.0
"""
salt.config.schemas.esxcluster
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

ESX Cluster configuration schemas
"""
from salt.utils.schema import AnyOfItem
from salt.utils.schema import ArrayItem
from salt.utils.schema import BooleanItem
from salt.utils.schema import ComplexSchemaItem
from salt.utils.schema import DefinitionsSchema
from salt.utils.schema import DictItem
from salt.utils.schema import IntegerItem
from salt.utils.schema import Schema
from salt.utils.schema import StringItem


class OptionValueItem(ComplexSchemaItem):
    """Sechma item of the OptionValue"""

    title = "OptionValue"
    key = StringItem(title="Key", required=True)
    value = AnyOfItem(items=[StringItem(), BooleanItem(), IntegerItem()])


class AdmissionControlPolicyItem(ComplexSchemaItem):
    """
    Schema item of the HA admission control policy
    """

    title = "Admission Control Policy"

    cpu_failover_percent = IntegerItem(title="CPU Failover Percent", minimum=0, maximum=100)
    memory_failover_percent = IntegerItem(title="Memory Failover Percent", minimum=0, maximum=100)


class DefaultVmSettingsItem(ComplexSchemaItem):
    """
    Schema item of the HA default vm settings
    """

    title = "Default VM Settings"

    isolation_response = StringItem(
        title="Isolation Response",
        enum=["clusterIsolationResponse", "none", "powerOff", "shutdown"],
    )
    restart_priority = StringItem(
        title="Restart Priority",
        enum=["clusterRestartPriority", "disabled", "high", "low", "medium"],
    )


class HAConfigItem(ComplexSchemaItem):
    """
    Schema item of ESX cluster high availability
    """

    title = "HA Configuration"
    description = "ESX cluster HA configuration json schema item"

    enabled = BooleanItem(title="Enabled", description="Specifies if HA should be enabled")
    admission_control_enabled = BooleanItem(title="Admission Control Enabled")
    admission_control_policy = AdmissionControlPolicyItem()
    default_vm_settings = DefaultVmSettingsItem()
    hb_ds_candidate_policy = StringItem(
        title="Heartbeat Datastore Candidate Policy",
        enum=["allFeasibleDs", "allFeasibleDsWithUserPreference", "userSelectedDs"],
    )
    host_monitoring = StringItem(title="Host Monitoring", choices=["enabled", "disabled"])
    options = ArrayItem(min_items=1, items=OptionValueItem())
    vm_monitoring = StringItem(
        title="Vm Monitoring",
        choices=["vmMonitoringDisabled", "vmAndAppMonitoring", "vmMonitoringOnly"],
    )


class VSANClusterConfigItem(ComplexSchemaItem):
    """
    Schema item of the ESX cluster vSAN configuration
    """

    title = "vSAN Configuration"
    description = "ESX cluster vSAN configurationi item"

    enabled = BooleanItem(title="Enabled", description="Specifies if vSAN should be enabled")
    auto_claim_storage = BooleanItem(
        title="Auto Claim Storage",
        description="Specifies whether the storage of member ESXi hosts should "
        "be automatically claimed for vSAN",
    )
    dedup_enabled = BooleanItem(title="Enabled", description="Specifies dedup should be enabled")
    compression_enabled = BooleanItem(
        title="Enabled", description="Specifies if compression should be enabled"
    )


class DRSConfigItem(ComplexSchemaItem):
    """
    Schema item of the ESX cluster DRS configuration
    """

    title = "DRS Configuration"
    description = "ESX cluster DRS configuration item"

    enabled = BooleanItem(title="Enabled", description="Specifies if DRS should be enabled")
    vmotion_rate = IntegerItem(
        title="vMotion rate",
        description="Aggressiveness to do automatic vMotions: "
        "1 (least aggressive) - 5 (most aggressive)",
        minimum=1,
        maximum=5,
    )
    default_vm_behavior = StringItem(
        title="Default VM DRS Behavior",
        description="Specifies the default VM DRS behavior",
        enum=["fullyAutomated", "partiallyAutomated", "manual"],
    )


class ESXClusterConfigSchema(DefinitionsSchema):
    """
    Schema of the ESX cluster config
    """

    title = "ESX Cluster Configuration Schema"
    description = "ESX cluster configuration schema"

    ha_item = HAConfigItem()
    vsan = VSANClusterConfigItem()
    drs = DRSConfigItem()
    vm_swap_placement = StringItem(title="VM Swap Placement")


class ESXClusterEntitySchema(Schema):
    """Schema of the ESX cluster entity"""

    title = "ESX Cluster Entity Schema"
    description = "ESX cluster entity schema"

    type = StringItem(
        title="Type",
        description="Specifies the entity type",
        required=True,
        enum=["cluster"],
    )

    datacenter = StringItem(
        title="Datacenter",
        description="Specifies the cluster datacenter",
        required=True,
        pattern=r"\w+",
    )

    cluster = StringItem(
        title="Cluster",
        description="Specifies the cluster name",
        required=True,
        pattern=r"\w+",
    )


class LicenseSchema(Schema):
    """
    Schema item of the ESX cluster vSAN configuration
    """

    title = "Licenses schema"
    description = "License configuration schema"

    licenses = DictItem(
        title="Licenses",
        description="Dictionary containing the license name to key mapping",
        required=True,
        additional_properties=StringItem(
            title="License Key",
            description="Specifies the license key",
            pattern=r"^(\w{5}-\w{5}-\w{5}-\w{5}-\w{5})$",
        ),
    )


class EsxclusterProxySchema(Schema):
    """
    Schema of the esxcluster proxy input
    """

    title = "Esxcluster Proxy Schema"
    description = "Esxcluster proxy schema"
    additional_properties = False
    proxytype = StringItem(required=True, enum=["esxcluster"])
    vcenter = StringItem(required=True, pattern=r"[^\s]+")
    datacenter = StringItem(required=True)
    cluster = StringItem(required=True)
    mechanism = StringItem(required=True, enum=["userpass", "sspi"])
    username = StringItem()
    passwords = ArrayItem(min_items=1, items=StringItem(), unique_items=True)
    # TODO Should be changed when anyOf is supported for schemas  # pylint: disable=W0511
    domain = StringItem()
    principal = StringItem()
    protocol = StringItem()
    port = IntegerItem(minimum=1)



"""
  Data required for create and update of the resource(security rule, Nat rule, Network etc..) on VMC
"""

create_sddc = {
    "account_link_config": None,
    "account_link_sddc_config": None,
    "deployment_type": "SingleAZ",
    "host_instance_type": "i3.metal",
    "msft_license_config": None,
    "name": "Salt-SDDC-1",
    "num_hosts": 0,
    "provider": "ZEROCLOUD",
    "sddc_id": None,
    "sddc_template_id": None,
    "sddc_type": None,
    "size": "medium",
    "skip_creating_vxlan": False,
    "sso_domain": "vmc.local",
    "storage_capacity": 0,
    "vpc_cidr": None,
    "vxlan_subnet": None,
    "region": "us-west-2",
}

create_sddc_cluster = {
    "host_cpu_cores_count": None,
    "host_instance_type": None,
    "msft_license_config": None,
    "num_hosts": 1,
    "storage_capacity": None,
}

manage_sddc_host = {
    "availability_zone": None,
    "cluster_id": None,
    "esxs": None,
    "num_hosts": 1,
    "strict_placement": False,
}


create_security_rules_mgw = {
    "sequence_number": 0,
    "source_groups": ["ANY"],
    "services": ["ANY"],
    "logged": False,
    "disabled": False,
    "destination_groups": ["ANY"],
    "scope": ["/infra/tier-1s/mgw"],
    "action": "ALLOW",
    "tag": "",
    "notes": "",
    "tags": None,
}


create_security_rules_cgw = {
    "sequence_number": 0,
    "source_groups": ["ANY"],
    "services": ["ANY"],
    "logged": False,
    "disabled": False,
    "destination_groups": ["ANY"],
    "scope": ["/infra/labels/cgw-all"],
    "action": "ALLOW",
    "tag": "",
    "notes": "",
    "tags": None,
}


update_security_rules = {
    "display_name": None,
    "sequence_number": None,
    "source_groups": None,
    "services": None,
    "logged": None,
    "disabled": None,
    "destination_groups": None,
    "scope": None,
    "action": None,
    "tag": None,
    "notes": None,
    "tags": None,
}


create_networks = {
    "subnets": None,
    "admin_state": "UP",
    "description": None,
    "domain_name": None,
    "tags": None,
    "advanced_config": None,
    "l2_extension": None,
    "dhcp_config_path": None,
}


update_networks = {
    "display_name": None,
    "subnets": None,
    "admin_state": None,
    "description": None,
    "domain_name": None,
    "tags": None,
    "advanced_config": None,
    "l2_extension": None,
    "dhcp_config_path": None,
}


create_dhcp_server_profiles = {
    "resource_type": "DhcpServerConfig",
    "server_addresses": None,
    "lease_time": None,
    "tags": None,
}


update_dhcp_server_profiles = {
    "display_name": None,
    "server_addresses": None,
    "server_address": None,
    "lease_time": None,
    "tags": None,
}


create_dhcp_relay_profiles = {
    "resource_type": "DhcpRelayConfig",
    "server_addresses": None,
    "tags": None,
}


update_dhcp_relay_profiles = {"display_name": None, "server_addresses": None, "tags": None}


create_distributed_firewall_rules = {
    "sequence_number": 1,
    "source_groups": ["ANY"],
    "destination_groups": ["ANY"],
    "scope": ["ANY"],
    "action": "DROP",
    "services": ["ANY"],
    "description": "",
    "disabled": False,
    "logged": False,
    "direction": "IN_OUT",
    "tag": "",
    "notes": "",
}

update_distributed_firewall_rules = {
    "source_groups": None,
    "destination_groups": None,
    "services": None,
    "scope": None,
    "action": None,
    "sequence_number": None,
    "display_name": None,
    "disabled": None,
    "logged": None,
    "description": None,
    "direction": None,
    "notes": None,
    "tag": None,
    "tags": None,
}


create_nat_rules = {
    "action": "REFLEXIVE",
    "description": "",
    "translated_network": None,
    "translated_ports": None,
    "destination_network": "",
    "source_network": None,
    "sequence_number": 0,
    "service": "",
    "logging": False,
    "enabled": True,
    "scope": ["/infra/labels/cgw-public"],
    "tags": None,
    "firewall_match": "MATCH_INTERNAL_ADDRESS",
}

update_nat_rules = {
    "action": None,
    "description": None,
    "translated_network": None,
    "translated_ports": None,
    "destination_network": None,
    "source_network": None,
    "sequence_number": None,
    "service": None,
    "logging": None,
    "enabled": None,
    "scope": None,
    "tags": None,
    "firewall_match": None,
    "display_name": None,
}


create_security_groups_cgw = {"expression": [], "tags": [], "description": ""}


create_security_groups_mgw = {"expression": [], "tags": [], "description": ""}

update_security_groups = {
    "expression": None,
    "tags": None,
    "display_name": None,
    "description": None,
}

create_vm_disks = {
    "backing": None,
    "ide": None,
    "new_vmdk": None,
    "sata": None,
    "scsi": None,
    "type": None,
}

add_org_users = {
    "skipNotify": False,
    "usernames": ["test@vmware.com"],
    "organizationRoles": [{"name": "org_member", "expiresAt": None}],
    "serviceRolesDtos": None,
    "customRoles": None,
    "skipNotifyRegistration": False,
    "invitedBy": None,
    "customGroupsIds": None,
}

remove_org_users = {"user_ids": None, "notify_users": False}



"""
Execution module to perform CRUD operations for NSX-T's Tier 1 Gateway
"""
import logging

from salt.exceptions import SaltInvocationError
from saltext.vmware.utils.nsxt_policy_base_resource import NSXTPolicyBaseResource
from saltext.vmware.utils.nsxt_resource_urls import DHCP_RELAY_CONFIG_URL
from saltext.vmware.utils.nsxt_resource_urls import EDGE_CLUSTER_URL
from saltext.vmware.utils.nsxt_resource_urls import EDGE_NODE_URL
from saltext.vmware.utils.nsxt_resource_urls import IPV6_DAD_PROFILE_URL
from saltext.vmware.utils.nsxt_resource_urls import IPV6_NDRA_PROFILE_URL
from saltext.vmware.utils.nsxt_resource_urls import SEGMENT_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_LOCALE_SERVICE_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_LS_INTERFACE_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_0_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_1_LOCALE_SERVICE_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_1_LS_INTERFACE_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_1_STATIC_ROUTE_URL
from saltext.vmware.utils.nsxt_resource_urls import TIER_1_URL

log = logging.getLogger(__name__)


def __virtual__():
    return "nsxt_policy_tier1"


class NSXTTier1(NSXTPolicyBaseResource):
    @classmethod
    def get_spec_identifier(cls):
        return "tier1"

    @staticmethod
    def get_resource_base_url(baseline_args=None):
        return TIER_1_URL

    @staticmethod
    def get_resource_base_query_params():
        return (
            "cursor",
            "include_mark_for_delete_objects",
            "included_fields",
            "page_size",
            "sort_ascending",
            "sort_by",
        )

    def update_resource_params(self, **kwargs):
        self.multi_resource_params = []
        fields = {
            "arp_limit",
            "default_rule_logging",
            "description",
            "disable_firewall",
            "display_name",
            "federation_config",
            "enable_standby_relocation",
            "failover_mode",
            "force_whitelisting",
            "intersite_config",
            "id",
            "pool_allocation",
            "qos_profile",
            "route_advertisement_rules",
            "route_advertisement_types",
            "state",
            "tags",
            "type",
            "_revision",
        }
        resource_params = {}
        for field in fields:
            if kwargs.get(field):
                resource_params[field] = kwargs[field]
        resource_params["resource_type"] = "Tier1"
        ipv6_profile_paths = []
        ipv6_ndra_profile_id = kwargs.get("ipv6_ndra_profile_id")
        if not ipv6_ndra_profile_id and kwargs.get("ipv6_ndra_profile_display_name"):
            ipv6_ndra_profile_id = self.get_id_using_display_name(
                url=(
                    NSXTTier1.get_nsxt_base_url().format(self.nsx_resource_params["hostname"])
                    + IPV6_NDRA_PROFILE_URL
                ),
                display_name=kwargs.get("ipv6_ndra_profile_display_name"),
            )
        if ipv6_ndra_profile_id:
            ipv6_profile_paths.append(IPV6_NDRA_PROFILE_URL + "/" + ipv6_ndra_profile_id)
        ipv6_dad_profile_id = kwargs.get("ipv6_dad_profile_id")
        if not ipv6_dad_profile_id and kwargs.get("ipv6_dad_profile_display_name"):
            ipv6_dad_profile_id = self.get_id_using_display_name(
                url=(
                    NSXTTier1.get_nsxt_base_url().format(self.nsx_resource_params["hostname"])
                    + IPV6_DAD_PROFILE_URL
                ),
                display_name=kwargs.get("ipv6_dad_profile_display_name"),
            )
        if ipv6_dad_profile_id:
            ipv6_profile_paths.append(IPV6_DAD_PROFILE_URL + "/" + ipv6_dad_profile_id)
        if ipv6_profile_paths:
            resource_params["ipv6_profile_paths"] = ipv6_profile_paths
        dhcp_config_id = kwargs.get("dhcp_config_id")
        if not dhcp_config_id and kwargs.get("dhcp_config_display_name"):
            dhcp_config_id = self.get_id_using_display_name(
                url=(
                    NSXTTier1.get_nsxt_base_url().format(self.nsx_resource_params["hostname"])
                    + DHCP_RELAY_CONFIG_URL
                ),
                display_name=kwargs.get("dhcp_config_display_name"),
            )
        if dhcp_config_id:
            resource_params["dhcp_config_paths"] = [DHCP_RELAY_CONFIG_URL + "/" + dhcp_config_id]
        tier0_id = kwargs.get("tier0_id")
        if not tier0_id and kwargs.get("tier0_display_name"):
            tier0_id = self.get_id_using_display_name(
                url=(
                    NSXTTier1.get_nsxt_base_url().format(self.nsx_resource_params["hostname"])
                    + TIER_0_URL
                ),
                display_name=kwargs.get("tier0_display_name"),
            )
        if tier0_id:
            resource_params["tier0_path"] = TIER_0_URL + "/" + tier0_id
        if not resource_params.get("id"):
            resource_params["id"] = resource_params["display_name"]
        self.multi_resource_params.append(resource_params)

    def update_parent_info(self, parent_info):
        parent_info["tier1_id"] = self.resource_params.get("id")

    class NSXTTier1StaticRoutes(NSXTPolicyBaseResource):
        def get_spec_identifier(self):
            return NSXTTier1.NSXTTier1StaticRoutes.get_spec_identifier()

        @classmethod
        def get_spec_identifier(cls):
            return "static_routes"

        def update_resource_params(self, **kwargs):
            self.multi_resource_params = []
            fields = {
                "description",
                "display_name",
                "enabled_on_secondary",
                "id",
                "network",
                "next_hops",
                "state",
                "tags",
                "_revision",
            }
            static_routes = kwargs.get("static_routes") or {}

            for static_route in static_routes:
                resource_params = {}
                for key in fields:
                    val = static_route.get(key)
                    if val:
                        resource_params[key] = val
                if not resource_params.get("id"):
                    resource_params["id"] = resource_params["display_name"]
                self.multi_resource_params.append(resource_params)

        @staticmethod
        def get_resource_base_url(parent_info):
            tier1_id = parent_info.get("tier1_id", "default")
            return TIER_1_STATIC_ROUTE_URL.format(tier1_id)

    class NSXTTier1LocaleService(NSXTPolicyBaseResource):
        def get_spec_identifier(self):
            return NSXTTier1.NSXTTier1LocaleService.get_spec_identifier()

        @classmethod
        def get_spec_identifier(cls):
            return "locale_services"

        @staticmethod
        def get_resource_base_url(parent_info):
            tier1_id = parent_info.get("tier1_id", "default")
            return TIER_1_LOCALE_SERVICE_URL.format(tier1_id)

        def update_resource_params(self, **kwargs):
            self.multi_resource_params = []
            fields = {
                "description",
                "display_name",
                "id",
                "route_redistribution_config",
                "state",
                "tags",
                "_revision",
            }
            locale_services = kwargs.get("locale_services") or {}
            base_url = NSXTPolicyBaseResource.get_nsxt_base_url().format(
                self.nsx_resource_params["hostname"]
            )
            for locale_service in locale_services:
                resource_params = {}
                for key in fields:
                    if locale_service.get(key):
                        resource_params[key] = locale_service.get(key)
                resource_params["resource_type"] = "LocaleServices"
                edge_cluster_info = locale_service.get("edge_cluster_info")
                if edge_cluster_info:
                    site_id = edge_cluster_info["site_id"]
                    enforcementpoint_id = edge_cluster_info["enforcementpoint_id"]
                    edge_cluster_base_url = EDGE_CLUSTER_URL.format(site_id, enforcementpoint_id)
                    edge_cluster_id = edge_cluster_info.get("edge_cluster_id")
                    if not edge_cluster_info and edge_cluster_info.get("edge_cluster_display_name"):
                        edge_cluster_id = self.get_id_using_display_name(
                            url=(base_url + edge_cluster_base_url),
                            display_name=edge_cluster_info.get("edge_cluster_display_name"),
                        )
                    if edge_cluster_id:
                        resource_params["edge_cluster_path"] = (
                            edge_cluster_base_url + "/" + edge_cluster_id
                        )
                preferred_edge_nodes_info = locale_service.get("preferred_edge_nodes_info")
                if preferred_edge_nodes_info:
                    resource_params["preferred_edge_paths"] = []
                    for preferred_edge_node_info in preferred_edge_nodes_info:
                        site_id = preferred_edge_node_info.get("site_id", "default")
                        enforcementpoint_id = preferred_edge_node_info.get(
                            "enforcementpoint_id", "default"
                        )
                        edge_cluster_base_url = EDGE_CLUSTER_URL.format(
                            site_id, enforcementpoint_id
                        )
                        edge_cluster_id = preferred_edge_node_info.get("edge_cluster_id")
                        if not edge_cluster_id and preferred_edge_node_info.get(
                            "edge_cluster_display_name"
                        ):
                            edge_cluster_id = self.get_id_using_display_name(
                                url=(
                                    NSXTTier1.get_nsxt_base_url().format(
                                        self.nsx_resource_params["hostname"]
                                    )
                                    + edge_cluster_base_url
                                ),
                                display_name=preferred_edge_node_info.get(
                                    "edge_cluster_display_name"
                                ),
                            )
                        if edge_cluster_id:
                            edge_node_base_url = EDGE_NODE_URL.format(
                                site_id, enforcementpoint_id, edge_cluster_id
                            )
                            edge_node_id = preferred_edge_node_info.get("edge_node_id")
                            if not edge_node_id and preferred_edge_node_info.get(
                                "edge_node_display_name"
                            ):
                                edge_node_id = self.get_id_using_display_name(
                                    url=(
                                        NSXTTier1.get_nsxt_base_url().format(
                                            self.nsx_resource_params["hostname"]
                                        )
                                        + edge_node_base_url
                                    ),
                                    display_name=preferred_edge_node_info.get(
                                        "edge_node_display_name"
                                    ),
                                )
                            if edge_node_id:
                                resource_params["preferred_edge_paths"].append(
                                    edge_node_base_url + "/" + edge_node_id
                                )
                bfd_profile_id = locale_service.get("bfd_profile_id")
                if not bfd_profile_id and locale_service.get("bfd_profile_display_name"):
                    bfd_profile_id = self.get_id_using_display_name(
                        url=(
                            NSXTTier1.get_nsxt_base_url().format(
                                self.nsx_resource_params["hostname"]
                            )
                            + "/infra/bfd-profiles"
                        ),
                        display_name=locale_service.get("bfd_profile_display_name"),
                    )
                if bfd_profile_id:
                    resource_params["bfd_profile_path"] = "/infra/bfd-profiles/{}".format(
                        bfd_profile_id
                    )
                ha_vip_configs = locale_service.get("ha_vip_configs")
                if ha_vip_configs:
                    resource_params["ha_vip_configs"] = []
                    for ha_vip_config in ha_vip_configs:
                        external_interface_info = ha_vip_config.pop("external_interface_info")
                        external_interface_paths = []
                        for external_interface in external_interface_info:
                            external_interface_path = external_interface.get(
                                "external_interface_path"
                            )
                            if not external_interface_path:
                                tier0_id = self.get_id_using_display_name(
                                    url=(
                                        NSXTTier1.get_nsxt_base_url().format(
                                            self.nsx_resource_params["hostname"]
                                        )
                                        + TIER_0_URL
                                    ),
                                    display_name=external_interface.get("tier0_display_name"),
                                )
                            tier0_ls_id = self.get_id_using_display_name(
                                url=(
                                    (
                                        NSXTTier1.get_nsxt_base_url().format(
                                            self.nsx_resource_params["hostname"]
                                        )
                                        + TIER_0_LOCALE_SERVICE_URL
                                    ).format(tier0_id)
                                ),
                                display_name=external_interface.get("locale_service_display_name"),
                            )
                            tier0_ls_inf_id = self.get_id_using_display_name(
                                url=(
                                    (
                                        NSXTTier1.get_nsxt_base_url().format(
                                            self.nsx_resource_params["hostname"]
                                        )
                                        + TIER_0_LS_INTERFACE_URL
                                    ).format(tier0_id, tier0_ls_id)
                                ),
                                display_name=external_interface.get("ls_interface_display_name"),
                            )
                            external_interface_path = (
                                TIER_0_LS_INTERFACE_URL.format(tier0_id, tier0_ls_id)
                                + "/"
                                + tier0_ls_inf_id
                            )
                            external_interface_paths.append(external_interface_path)
                        ha_vip_config["external_interface_paths"] = external_interface_paths
                        ha_vip_config["vip_subnets"] = ha_vip_config.pop("vip_subnets")
                        resource_params["ha_vip_configs"].append(ha_vip_config)
                if not resource_params.get("id"):
                    resource_params["id"] = resource_params["display_name"]
                self.multi_resource_params.append(resource_params)

        def update_parent_info(self, parent_info):
            parent_info["locale_services_id"] = self.resource_params["id"]
            parent_info["ls_display_name"] = self.resource_params["display_name"]

        class NSXTTier1Interface(NSXTPolicyBaseResource):
            def get_spec_identifier(self):
                return NSXTTier1.NSXTTier1LocaleService.NSXTTier1Interface.get_spec_identifier()

            @classmethod
            def get_spec_identifier(cls):
                return "interfaces"

            @staticmethod
            def get_resource_base_url(parent_info):
                tier1_id = parent_info.get("tier1_id", "default")
                locale_service_id = parent_info.get("locale_services_id", "default")
                return TIER_1_LS_INTERFACE_URL.format(tier1_id, locale_service_id)

            def update_resource_params(self, **kwargs):
                self.multi_resource_params = []
                fields = {
                    "description",
                    "display_name",
                    "id",
                    "mtu",
                    "state",
                    "subnets",
                    "tags",
                    "urpf_mode",
                    "_revision",
                }
                locale_services = kwargs.get("locale_services") or {}
                ls_display_name = self._parent_info.get("ls_display_name")
                locale_service = next(
                    (ls for ls in locale_services if ls.get("display_name") == ls_display_name),
                    None,
                )
                if locale_service:
                    interfaces = locale_service.get("interfaces") or {}
                    for interface in interfaces:
                        resource_params = {}
                        resource_params["resource_type"] = "Tier1Interface"
                        for field in fields:
                            val = interface.get(field)
                            if val:
                                resource_params[field] = val
                        # segment_id is a required attr
                        segment_id = interface.get("segment_id")
                        if not segment_id and interface.get("segment_display_name"):
                            segment_id = self.get_id_using_display_name(
                                url=(
                                    NSXTTier1.get_nsxt_base_url().format(
                                        self.nsx_resource_params["hostname"]
                                    )
                                    + SEGMENT_URL
                                ),
                                display_name=interface.get("segment_display_name"),
                            )
                        if segment_id:
                            resource_params["segment_path"] = SEGMENT_URL + "/" + segment_id
                        ipv6_ndra_profile_id = interface.get("ipv6_ndra_profile_id")
                        ipv6_ndra_profile_display_name = interface.get(
                            "ipv6_ndra_profile_display_name"
                        )
                        if not ipv6_ndra_profile_id and ipv6_ndra_profile_display_name:
                            ipv6_ndra_profile_id = self.get_id_using_display_name(
                                url=(
                                    NSXTTier1.get_nsxt_base_url().format(
                                        self.nsx_resource_params["hostname"]
                                    )
                                    + IPV6_NDRA_PROFILE_URL
                                ),
                                display_name=ipv6_ndra_profile_display_name,
                            )
                        if ipv6_ndra_profile_id:
                            resource_params["ipv6_profile_paths"] = [
                                IPV6_NDRA_PROFILE_URL + "/" + ipv6_ndra_profile_id
                            ]
                        dhcp_config_id = interface.get("segment_id")
                        if not dhcp_config_id and interface.get("dhcp_config_display_name"):
                            dhcp_config_id = self.get_id_using_display_name(
                                url=(
                                    NSXTTier1.get_nsxt_base_url().format(
                                        self.nsx_resource_params["hostname"]
                                    )
                                    + DHCP_RELAY_CONFIG_URL
                                ),
                                display_name=interface.get("dhcp_config_display_name"),
                            )
                        if dhcp_config_id:
                            resource_params["dhcp_relay_path"] = (
                                DHCP_RELAY_CONFIG_URL + "/" + dhcp_config_id
                            )
                        if not resource_params.get("id"):
                            resource_params["id"] = resource_params["display_name"]
                        self.multi_resource_params.append(resource_params)


def get_by_display_name(
    hostname, username, password, display_name, verify_ssl=True, cert=None, cert_common_name=None
):
    """
    Gets Tier 1 Gateway present in the NSX-T Manager with given name.

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_policy_tier1.get_by_display_name hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    display_name
        The name of tier 1 gateway to fetch

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """
    nsxt_tier1 = NSXTTier1()
    url = (NSXTPolicyBaseResource.get_nsxt_base_url() + nsxt_tier1.get_resource_base_url()).format(
        hostname
    )
    return nsxt_tier1.get_by_display_name(
        url,
        username,
        password,
        display_name,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
    )


def get(
    hostname,
    username,
    password,
    verify_ssl=True,
    cert=None,
    cert_common_name=None,
    cursor=None,
    included_fields=None,
    page_size=None,
    sort_ascending=None,
    sort_by=None,
):
    """
    Lists NSXT Tier 1 Gateways present in the NSX-T Manager


    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_policy_tier1.get hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    cursor
        (Optional) Opaque cursor to be used for getting next page of records (supplied by current result page)
        include_mark_for_delete_objects

        (Optional) Include objects that are marked for deletion in results. If true, resources that are marked for
        deletion will be included in the results. By default, these resources are not included.

    included_fields
        (Optional) Comma separated list of fields that should be included in query result

    page_size
        (Optional) Maximum number of results to return in this page

    sort_by
        (Optional) Field by which records are sorted

    sort_ascending
        (Optional) Boolean value to sort result in ascending order

    """
    nsxt_tier1 = NSXTTier1()
    url = (NSXTPolicyBaseResource.get_nsxt_base_url() + nsxt_tier1.get_resource_base_url()).format(
        hostname
    )
    return nsxt_tier1.get(
        url,
        username,
        password,
        verify_ssl=verify_ssl,
        cert=cert,
        cert_common_name=cert_common_name,
        cursor=cursor,
        included_fields=included_fields,
        page_size=page_size,
        sort_ascending=sort_ascending,
        sort_by=sort_by,
    )


def create_or_update(
    hostname,
    username,
    password,
    cert=None,
    cert_common_name=None,
    verify_ssl=True,
    display_name=None,
    tags=None,
    id=None,
    description=None,
    default_rule_logging=None,
    disable_firewall=None,
    failover_mode=None,
    enable_standby_relocation=None,
    force_whitelisting=None,
    intersite_config=None,
    ipv6_ndra_profile_id=None,
    ipv6_ndra_profile_display_name=None,
    ipv6_dad_profile_id=None,
    ipv6_dad_profile_display_name=None,
    dhcp_config_id=None,
    dhcp_config_display_name=None,
    pool_allocation=None,
    qos_profile=None,
    route_advertisement_rules=None,
    route_advertisement_types=None,
    tier0_id=None,
    tier0_display_name=None,
    static_routes=None,
    locale_services=None,
    arp_limit=None,
    type=None,
):
    """
    Creates a Tier 1 Gateway and its sub-resources with given specifications

    CLI Example:

    .. code-block:: bash

        salt vm_minion nsxt_policy_tier1.create hostname=nsxt-manager.local username=admin ...

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    display_name:

        description:
            - Display name.
            - If resource ID is not specified, display_name will be used as ID.

        required: true
        type: str

    state:
        description: present or absent keyword is used as an indetifier, default value is present.
                    If a user has provided absent that resource/sub-resource will be deleted

    tags:
        description: Opaque identifiers meaningful to the API user.
        type: dict
        suboptions:

            scope:
                description: Tag scope.
                required: true
                type: str

            tag:
                description: Tag value.
                required: true
                type: str

    id:
        description: Tier-1 ID
        required: false
        type: str

    description:
        description: Tier-1 description
        type: str

    default_rule_logging:
        description: Enable logging for whitelisted rule.
                     Indicates if logging should be enabled for the default
                     whitelisting rule.

        default: false

    disable_firewall:
        description: Disable or enable gateway fiewall.

        default: False
        type: bool

    failover_mode:

        description: Determines the behavior when a Tier-1 instance in
                     ACTIVE-STANDBY high-availability mode restarts
                     after a failure. If set to PREEMPTIVE, the preferred node
                     will take over, even if it causes
                     another failure. If set to NON_PREEMPTIVE, then
                     the instance that restarted will remain secondary.
                     This property must not be populated unless the
                     ha_mode property is set to ACTIVE_STANDBY.

        choices:

            - 'NON_PREEMPTIVE'
            - 'PREEMPTIVE'

        type: str

    enable_standby_relocation:

        description:
            - Flag to enable standby service router relocation.
            - Standby relocation is not enabled until edge cluster is
              configured for Tier1.

        type: bool
        default: false

    force_whitelisting:

        description: Flag to add whitelisting FW rule during
                     realization.

        default: False
        type: bool

    intersite_config:

        description: Inter site routing configuration when the gateway is
                     streched.

        type: dict
        suboptions:

            fallback_sites:

                description: Fallback site to be used as new primary
                             site on current primary site failure.
                             Disaster recovery must be initiated via
                             API/UI. Fallback site configuration is
                             supported only for T0 gateway. T1 gateway
                             will follow T0 gateway's primary site
                             during disaster recovery.

                type: list

            intersite_transit_subnet:

                description:
                    - Transit subnet in CIDR format
                    - IPv4 subnet for inter-site transit segment
                      connecting service routers across sites for
                      stretched gateway. For IPv6 link local subnet is
                      auto configured

                type: str
                default: "169.254.32.0/20"

            last_admin_active_epoch:

                description:
                    - Epoch of last time admin changing active
                      LocaleServices
                    - Epoch(in seconds) is auto updated based on
                      system current timestamp when primary locale
                      service is updated. It is used for resolving
                      conflict during site failover. If system clock
                      not in sync then User can optionally override
                      this. New value must be higher than the current
                      value.

                type: int

            primary_site_path:

                description:
                    - Primary egress site for gateway.
                    - Primary egress site for gateway. T0/T1 gateway in
                      Active/Standby mode supports stateful services on primary
                      site. In this mode primary site must be set if gateway is
                      stretched to more than one site. For T0 gateway in
                      Active/Active primary site is optional field. If set then
                      secondary site prefers routes learned from primary over
                      locally learned routes. This field is not applicable for
                      T1 gateway with no services

                type: str

    ipv6_ndra_profile_id:

        description: IPv6 NDRA profile configuration on Tier1.
                     Either or both NDRA and/or DAD profiles can be
                     configured. Related attribute ipv6_dad_profile_id.

        type: str

    ipv6_ndra_profile_display_name:

        description: Same as ipv6_ndra_profile_id. Either one can be specified.
                     If both are specified, ipv6_ndra_profile_id takes
                     precedence.

        type: str

    ipv6_dad_profile_id:

        description: IPv6 DRA profile configuration on Tier1.
                     Either or both NDRA and/or DAD profiles can be
                     configured. Related attribute ipv6_ndra_profile_id.

        type: str

    ipv6_dad_profile_display_name:

        description: Same as ipv6_dad_profile_id. Either one can be specified.
                     If both are specified, ipv6_dad_profile_id takes
                     precedence.

        type: str

    dhcp_config_id:

        description: DHCP configuration for Segments connected to
                     Tier-1. DHCP service is configured in relay mode.

        type: str

    dhcp_config_display_name:

        description: Same as dhcp_config_id. Either one can be specified.
                     If both are specified, dhcp_config_id takes precedence.

        type: str

    pool_allocation:

        description:
            - Edge node allocation size
            - Supports edge node allocation at different sizes for routing and
              load balancer service to meet performance and scalability
              requirements.
            - ROUTING - Allocate edge node to provide routing services.
            - LB_SMALL, LB_MEDIUM, LB_LARGE, LB_XLARGE - Specify size of load
              balancer service that will be configured on TIER1 gateway.

        type: str

        choices:

            - ROUTING
            - LB_SMALL
            - LB_MEDIUM
            - LB_LARGE
            - LB_XLARGE

        default: ROUTING

    qos_profile:

        description: QoS Profile configuration for Tier1 router link connected
                     to Tier0 gateway.

        type: dict

        suboptions:

            egress_qos_profile_path:
                description: Policy path to gateway QoS profile in egress
                             direction.

                type: str

            ingress_qos_profile_path:
                description: Policy path to gateway QoS profile in ingress
                             direction.

                type: str

    route_advertisement_rules:

        description: Route advertisement rules and filtering

        type: list

        suboptions:

            action:
                description:
                    - Action to advertise filtered routes to the connected
                      Tier0 gateway.
                choices:
                    - PERMIT: Enables the advertisment
                    - DENY: Disables the advertisement

                type: str
                required: true

            name:
                description: Display name for rule
                type: str
                required: true

            prefix_operator:
                description:
                    - Prefix operator to filter subnets.
                    - GE prefix operator filters all the routes with prefix
                      length greater than or equal to the subnets configured.
                    - EQ prefix operator filter all the routes with prefix
                      length equal to the subnets configured.

                type: str

                choices:

                    - GE
                    - EQ

            route_advertisement_types:

                description:
                    - Enable different types of route advertisements.
                    - By default, Routes to IPSec VPN local-endpoint subnets
                      (TIER1_IPSEC_LOCAL_ENDPOINT) are advertised if no value
                      is supplied here.

                type: list
                choices:

                    - 'TIER1_STATIC_ROUTES'
                    - 'TIER1_CONNECTED'
                    - 'TIER1_NAT'
                    - 'TIER1_LB_VIP'
                    - 'TIER1_LB_SNAT'
                    - 'TIER1_DNS_FORWARDER_IP'
                    - 'TIER1_IPSEC_LOCAL_ENDPOINT'

            subnets:
                description: Network CIDRs to be routed.
                type: list

    route_advertisement_types:

        description:
            - Enable different types of route advertisements.
            - By default, Routes to IPSec VPN local-endpoint subnets
              (TIER1_IPSEC_LOCAL_ENDPOINT) are advertised if no value is
              supplied here.

        type: list
        choices:

            - 'TIER1_STATIC_ROUTES'
            - 'TIER1_CONNECTED'
            - 'TIER1_NAT'
            - 'TIER1_LB_VIP'
            - 'TIER1_LB_SNAT'
            - 'TIER1_DNS_FORWARDER_IP'
            - 'TIER1_IPSEC_LOCAL_ENDPOINT'

    tier0_id:

        description: Tier-1 connectivity to Tier-0
        type: str

    tier0_display_name:

        description: Same as tier0_id. Either one can be specified.
                    If both are specified, tier0_id takes precedence.

        type: str

    static_routes:

        type: list
        element: dict
        description: This is a list of Static Routes that need to be created,updated, or deleted

        suboptions:

            id:
                description: Tier-1 Static Route ID.
                required: false

                type: str

            display_name:
                description:
                    - Tier-1 Static Route display name.
                    - Either this or id must be specified. If both are
                      specified, id takes precedence.

                required: true
                type: str

            description:

                description:
                    - Tier-1 Static Route description.

                type: str

            state:

                description:
                    - State can be either 'present' or 'absent'. 'present' is
                      used to create or update resource. 'absent' is used to
                      delete resource.
                    - Must be specified in order to modify the resource

                choices:

                    - present
                    - absent

            network:
                description: Network address in CIDR format
                required: true
                type: str

            next_hops:
                description: Next hop routes for network
                type: list
                elements: dict

                suboptions:

                    admin_distance:
                        description: Cost associated with next hop route
                        type: int
                        default: 1

                ip_address:
                    description: Next hop gateway IP address
                    type: str

                scope:
                    description:
                        - Interface path associated with current route
                        - For example, specify a policy path referencing the
                          IPSec VPN Session

                    type: list

            tags:
                description: Opaque identifiers meaningful to the API user
                type: dict
                suboptions:

                    scope:

                        description: Tag scope.
                        required: true
                        type: str

                    tag:
                        description: Tag value.
                        required: true
                        type: str

    locale_services:
        type: list
        element: dict
        description: This is a list of Locale Services that need to be created,updated, or deleted

        suboptions:

            id:
                description: Tier-1 Locale Service ID
                type: str

            display_name:
                description:
                    - Tier-1 Locale Service display name.
                    - Either this or id must be specified. If both are
                      specified, id takes precedence.

                required: true
                type: str

            description:

                description: Tier-1 Locale Service  description
                type: str

            state:

                description:
                    - State can be either 'present' or 'absent'. 'present' is
                      used to create or update resource. 'absent' is used to
                      delete resource.
                    - Required if I(segp_id != null)

                choices:

                    - present
                    - absent

            tags:
                description: Opaque identifiers meaningful to the API user.
                type: dict
                suboptions:

                    scope:

                        description: Tag scope.
                        required: true
                        type: str

                    tag:
                        description: Tag value.
                        required: true
                        type: str

            edge_cluster_info:
                description: Used to create path to edge cluster. Auto-assigned
                             if associated enforcement-point has only one edge
                             cluster.

                type: dict
                suboptions:

                    site_id:

                        description: site_id where edge cluster is located
                        default: default
                        type: str

                    enforcementpoint_id:
                        description: enforcementpoint_id where edge cluster is
                                     located

                        default: default
                        type: str

                    edge_cluster_id:
                        description: ID of the edge cluster

                        required: true
                        type: str

                    edge_cluster_display_name:
                        description:
                            - display name of the edge cluster.
                            - Either this or edge_cluster_id must be specified.
                              If both are specified, edge_cluster_id takes
                              precedence

                        type: str

            preferred_edge_nodes_info:

                description: Used to create paths to edge nodes. Specified edge
                             is used as preferred edge cluster member when
                             failover mode is set to PREEMPTIVE, not
                             applicable otherwise.

                type: list
                suboptions:

                    site_id:

                        description: site_id where edge node is located
                        default: default
                        type: str

                    enforcementpoint_id:

                        description: enforcementpoint_id where edge node is
                                     located

                        default: default
                        type: str

                    edge_cluster_id:
                        description: edge_cluster_id where edge node is
                                     located

                        required: true
                        type: str

                    edge_cluster_display_name:

                        description:
                            - display name of the edge cluster.
                            - either this or edge_cluster_id must be specified.
                              If both are specified, edge_cluster_id takes
                              precedence

                        type: str

                    edge_node_id:

                        description: ID of the edge node
                        type: str

                    edge_node_display_name:

                        description:
                            - Display name of the edge node.
                            - either this or edge_node_id must be specified. If
                              both are specified, edge_node_id takes precedence

                        type: str

            route_redistribution_types:
                description:
                    - Enable redistribution of different types of routes on
                      Tier-0.
                    - This property is only valid for locale-service under
                      Tier-0.
                    - This property is deprecated, please use
                      "route_redistribution_config" property to configure
                      redistribution rules.

                choices:

                    - TIER0_STATIC - Redistribute user added
                        static routes.
                    - TIER0_CONNECTED - Redistribute all
                        subnets configured on Interfaces and
                        routes related to TIER0_ROUTER_LINK,
                        TIER0_SEGMENT, TIER0_DNS_FORWARDER_IP,
                        TIER0_IPSEC_LOCAL_IP, TIER0_NAT types.
                    - TIER1_STATIC - Redistribute all subnets
                        and static routes advertised by Tier-1s.
                    - TIER0_EXTERNAL_INTERFACE - Redistribute
                        external interface subnets on Tier-0.
                    - TIER0_LOOPBACK_INTERFACE - Redistribute
                        loopback interface subnets on Tier-0.
                    - TIER0_SEGMENT - Redistribute subnets
                        configured on Segments connected to
                        Tier-0.
                    - TIER0_ROUTER_LINK - Redistribute router
                        link port subnets on Tier-0.
                    - TIER0_SERVICE_INTERFACE - Redistribute
                        Tier0 service interface subnets.
                    - TIER0_DNS_FORWARDER_IP - Redistribute DNS
                        forwarder subnets.
                    - TIER0_IPSEC_LOCAL_IP - Redistribute IPSec
                        subnets.
                    - TIER0_NAT - Redistribute NAT IPs owned by
                        Tier-0.
                    - TIER0_EVPN_TEP_IP - Redistribute EVPN
                        local endpoint subnets on Tier-0.
                    - TIER1_NAT - Redistribute NAT IPs
                        advertised by Tier-1 instances.
                    - TIER1_LB_VIP - Redistribute LB VIP IPs
                        advertised by Tier-1 instances.
                    - TIER1_LB_SNAT - Redistribute LB SNAT IPs
                        advertised by Tier-1 instances.
                    - TIER1_DNS_FORWARDER_IP - Redistribute DNS
                        forwarder subnets on Tier-1 instances.
                    - TIER1_CONNECTED - Redistribute all
                        subnets configured on Segments and
                        Service Interfaces.
                    - TIER1_SERVICE_INTERFACE - Redistribute
                        Tier1 service interface subnets.
                    - TIER1_SEGMENT - Redistribute subnets
                        configured on Segments connected to
                        Tier1.
                    - TIER1_IPSEC_LOCAL_ENDPOINT - Redistribute
                        IPSec VPN local-endpoint subnets advertised by TIER1.

                type: list

            route_redistribution_config:

                description: Configure all route redistribution properties like
                             enable/disable redistributon, redistribution rule
                             and so on.

                type: dict

                suboptions:

                    bgp_enabled:

                        description: Flag to enable route redistribution.
                        type: bool
                        default: false

                    redistribution_rules:

                        description: List of redistribution rules.
                        type: list
                        elements: dict

                        suboptions:

                            name:
                                description: Rule name
                                type: str

                            route_map_path:
                                description: Route map to be associated with
                                             the redistribution rule

                                type: str

                            route_redistribution_types:
                                description: Tier-0 route redistribution types

                                choices:

                                    - TIER0_STATIC - Redistribute user added
                                      static routes.
                                    - TIER0_CONNECTED - Redistribute all
                                      subnets configured on Interfaces and
                                      routes related to TIER0_ROUTER_LINK,
                                      TIER0_SEGMENT, TIER0_DNS_FORWARDER_IP,
                                      TIER0_IPSEC_LOCAL_IP, TIER0_NAT types.
                                    - TIER1_STATIC - Redistribute all subnets
                                      and static routes advertised by Tier-1s.
                                    - TIER0_EXTERNAL_INTERFACE - Redistribute
                                      external interface subnets on Tier-0.
                                    - TIER0_LOOPBACK_INTERFACE - Redistribute
                                      loopback interface subnets on Tier-0.
                                    - TIER0_SEGMENT - Redistribute subnets
                                      configured on Segments connected to
                                      Tier-0.
                                    - TIER0_ROUTER_LINK - Redistribute router
                                      link port subnets on Tier-0.
                                    - TIER0_SERVICE_INTERFACE - Redistribute
                                      Tier0 service interface subnets.
                                    - TIER0_DNS_FORWARDER_IP - Redistribute DNS
                                      forwarder subnets.
                                    - TIER0_IPSEC_LOCAL_IP - Redistribute IPSec
                                      subnets.
                                    - TIER0_NAT - Redistribute NAT IPs owned by
                                      Tier-0.
                                    - TIER0_EVPN_TEP_IP - Redistribute EVPN
                                      local endpoint subnets on Tier-0.
                                    - TIER1_NAT - Redistribute NAT IPs
                                      advertised by Tier-1 instances.
                                    - TIER1_LB_VIP - Redistribute LB VIP IPs
                                      advertised by Tier-1 instances.
                                    - TIER1_LB_SNAT - Redistribute LB SNAT IPs
                                      advertised by Tier-1 instances.
                                    - TIER1_DNS_FORWARDER_IP - Redistribute DNS
                                      forwarder subnets on Tier-1 instances.
                                    - TIER1_CONNECTED - Redistribute all
                                      subnets configured on Segments and
                                      Service Interfaces.
                                    - TIER1_SERVICE_INTERFACE - Redistribute
                                      Tier1 service interface subnets.
                                    - TIER1_SEGMENT - Redistribute subnets
                                      configured on Segments connected to
                                      Tier1.
                                    - TIER1_IPSEC_LOCAL_ENDPOINT - Redistribute
                                      IPSec VPN local-endpoint subnets
                                      advertised by TIER1.

                                type: list

            ha_vip_configs:
                type: list
                elements: dict
                description:

                    - Array of HA VIP Config.
                    - This configuration can be defined only for Active-Standby
                      Tier0 gateway to provide redundancy. For mulitple
                      external interfaces, multiple HA VIP configs must be
                      defined and each config will pair exactly two external
                      interfaces. The VIP will move and will always be owned by
                      the Active node. When this property is configured,
                      configuration of dynamic-routing is not allowed.

                suboptions:

                    enabled:

                        description: Flag to enable this HA VIP config.
                        default: true
                        type: bool

                    external_interface_info:
                        type: list
                        elements: dict
                        description: Array of external interface info
                        external_interface_paths:

                            description:
                                - Policy paths to Tier0 external interfaces for
                                  providing redundancy
                                - Policy paths to Tier0 external interfaces which
                                  are to be paired to provide redundancy. Floating
                                  IP will be owned by one of these interfaces
                                  depending upon which edge node is Active.
                                - minimum 2 values should be present

                            type: list

                        tier0_display_name:

                            description: tier0 display name to create the external
                                         interface paths. Can be skipped if external
                                         interface paths are provided

                        locale-service_display_name:
                            description: locale-service display name attached to the provided
                                         tier0. Can be skipped if external interface paths are provided

                        ls_interface_display_name:
                            description: interface attached to the provided tier0 and locale-service
                                         Can be skipped if external interface paths are provided

                    vip_subnets:

                        description:
                            - VIP floating IP address subnets
                            - Array of IP address subnets which will be used as
                              floating IP addresses.

                        type: list

                        suboptions:

                            ip_addresses:

                                description: IP addresses assigned to interface
                                type: list
                                required: true

                            prefix_len:
                                description: Subnet prefix length
                                type: int
                                required: true

            interfaces:
                type: list
                element: dict

                description: Specify the interfaces associated with the Gateway

                suboptions:

                    id:
                        description: Tier-1 Interface ID
                        required: false
                        type: str

                    description:
                        description: Tier-1 Interface  description
                        type: str

                    display_name:
                        description:
                            - Tier-1 Interface display name
                            - Either this or id must be specified. If both are
                              specified, id takes precedence.

                        required: false
                        type: str

                    state:
                        description:
                            - State can be either 'present' or 'absent'.
                              'present' is used to create or update resource.
                              'absent' is used to delete resource.
                            - Required if I(segp_id != null).

                        choices:

                            - present
                            - absent

                    tags:

                        description: Opaque identifiers meaningful to the API
                                     user

                        type: dict
                        suboptions:

                            scope:
                                description: Tag scope.
                                required: true
                                type: str

                            tag:
                                description: Tag value.
                                required: true
                                type: str

                    ipv6_ndra_profile_id:

                        description:
                            - Configrue IPv6 NDRA profile. Only one NDRA
                              profile can be configured
                            - Required if I(id != null)

                        type: str

                    mtu:
                        description:
                            - MTU size
                            - Maximum transmission unit (MTU) specifies the
                              size of the largest packet that a network
                              protocol can transmit.

                        type: int

                    segment_id:

                        description:
                            - Specify Segment to which this interface is
                              connected to.
                            - Required if I(id != null)

                        type: str

                    segment_display_name:

                        description:
                            - Same as segment_id
                            - Either this or segment_id must be specified. If
                              both are specified, segment_id takes precedence.

                        type: str

                    subnets:

                        description:
                            - IP address and subnet specification for interface
                            - Specify IP address and network prefix for
                              interface
                            - Required if I(id != null)

                        type: list
                        elements: dict

                        suboptions:

                            ip_addresses:
                                description: IP addresses assigned to interface
                                type: str

                            prefix_len:
                                description: Subnet prefix length
                                type: str

                    dhcp_config_id:
                        description: id of referenced dhcp-relay-config
                        type: str

                    dhcp_display_name:
                        description: name of referenced dhcp-relay-config

                    urpf_mode:
                        description: Unicast Reverse Path Forwarding mode
                        type: str
                        requires: False
                        choices:

                            - NONE
                            - STRICT

                        default: STRICT
    """
    execution_logs = []
    nsxt_tier1 = NSXTTier1()
    try:
        nsxt_tier1.create_or_update(
            hostname=hostname,
            username=username,
            password=password,
            execution_logs=execution_logs,
            cert=cert,
            cert_common_name=cert_common_name,
            verify_ssl=verify_ssl,
            display_name=display_name,
            tags=tags,
            id=id,
            description=description,
            default_rule_logging=default_rule_logging,
            disable_firewall=disable_firewall,
            failover_mode=failover_mode,
            enable_standby_relocation=enable_standby_relocation,
            force_whitelisting=force_whitelisting,
            intersite_config=intersite_config,
            ipv6_ndra_profile_id=ipv6_ndra_profile_id,
            ipv6_ndra_profile_display_name=ipv6_ndra_profile_display_name,
            ipv6_dad_profile_id=ipv6_dad_profile_id,
            ipv6_dad_profile_display_name=ipv6_dad_profile_display_name,
            dhcp_config_id=dhcp_config_id,
            dhcp_config_display_name=dhcp_config_display_name,
            pool_allocation=pool_allocation,
            qos_profile=qos_profile,
            route_advertisement_rules=route_advertisement_rules,
            route_advertisement_types=route_advertisement_types,
            tier0_id=tier0_id,
            tier0_display_name=tier0_display_name,
            static_routes=static_routes,
            locale_services=locale_services,
            arp_limit=arp_limit,
            type=type,
        )
    except SaltInvocationError as e:
        execution_logs.append({"error": str(e)})
    return execution_logs


def delete(
    hostname, username, password, tier1_id, cert=None, cert_common_name=None, verify_ssl=True
):
    """
    Deletes a Tier 1 gateway and it sub-resources

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    tier1_id
        id of the tier 1 to be deleted

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name
        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """
    execution_logs = []
    nsxt_tier1 = NSXTTier1()
    try:
        nsxt_tier1.delete(
            hostname,
            username,
            password,
            tier1_id,
            cert,
            cert_common_name,
            verify_ssl,
            execution_logs,
        )
    except SaltInvocationError as e:
        execution_logs.append({"error": str(e)})
    return execution_logs


def get_hierarchy(
    hostname, username, password, tier1_id, cert=None, cert_common_name=None, verify_ssl=True
):
    """
    Returns entire hierarchy of Tier 1 gateway and its sub-resources

    hostname
        The host name of NSX-T manager

    username
        Username to connect to NSX-T manager

    password
        Password to connect to NSX-T manager

    tier1_id
        id of the tier 1 gateway

    verify_ssl
        Option to enable/disable SSL verification. Enabled by default.
        If set to False, the certificate validation is skipped.

    cert
        (Optional) Path to the SSL client certificate file to connect to NSX-T manager.
        The certificate can be retrieved from browser.

    cert_common_name

        (Optional) By default, the hostname parameter and the common name in certificate is compared for host name
        verification. If the client certificate common name and hostname do not match (in case of self-signed
        certificates), specify the certificate common name as part of this parameter. This value is then used to
        compare against

    """
    result = {}
    nsxt_tier1 = NSXTTier1()
    try:
        nsxt_tier1.get_hierarchy(
            hostname, username, password, tier1_id, cert, cert_common_name, verify_ssl, result
        )
        log.info("Hierarchy result for tier 1 gateway: {}".format(result))
        return result
    except SaltInvocationError as e:
        return {"error": str(e)}



# Copyright 2022 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.esxi as utils_esxi

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim, vmodl

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

__virtualname__ = "vmware_vswitch"
__proxyenabled__ = ["vmware_vswitch"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def _get(hostname, switch_name=None, service_instance=None, profile=None):
    """
    Returns a list of vswitches found on a host; if switch_name is set, the
    list will contain just that vswitch.
    If no matching vswitches are found, the empty list is returned.

    hostname
        The host name to inspect.

    switch_name
        The vswitch name (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    host = utils_esxi.get_host(hostname, service_instance)

    vswitches = host.configManager.networkSystem.networkInfo.vswitch

    if switch_name:
        for vswitch in vswitches:
            if vswitch.name == switch_name:
                return [vswitch]
        return []

    return vswitches


def get(hostname, switch_name=None, service_instance=None, profile=None):
    """
    Get the properties of all vswitches on a host, optionally filtering to one name.
    Returns an empty list if no matching vswitch is found.

    hostname
        The host name to inspect.

    switch_name
        The vswitch name (optional).

    service_instance
        Use this vCenter service connection instance instead of creating a new one (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vswitch.get hostname=host1 switch_name=vSwitch1
    """
    ret = []
    for vswitch in _get(
        hostname=hostname,
        switch_name=switch_name,
        service_instance=service_instance,
        profile=profile,
    ):
        spec = vswitch.spec
        info = {
            "num_ports": spec.numPorts,
            "mtu": spec.mtu or vswitch.mtu,
            "name": vswitch.name,
            # each nic looks something like "key-vim.host.PhysicalNic-vmnic0"
            "nics": [nic.split("-")[-1] for nic in vswitch.pnic],
            # TODO: unit test "key-vim.host.PortGroup-Management Network", only split twice in case '-' is in the name
            "portgroups": [pg.split("-", 2)[-1] for pg in vswitch.portgroup],
            # TODO: support bonded/bridged nics
            # TODO: support "policy", teaming, offload, security and shaping.
        }
        ret.append(info)

    return ret


def add(
    switch_name, hostname, mtu=1500, nics=[], num_ports=128, service_instance=None, profile=None
):
    """
    Adds a vswitch to the host.

    switch_name
        The name of the new vswitch.

    hostname
        The hostname where the switch will be added.

    mtu
        Maximum transmission unit size in bytes (optional).

    nics
        List of physical nics to attach to; if used might look like ["vmnic0"] (optional).

    num_ports
        Number of ports to allocate on the virtual switch.

    service_instance
        Use this vCenter service connection instance instead of creating a new one (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vswitch.add switch_name=vSwitch0 hostname=host1 mtu=1500, nics='["vmnic0", "vmnic1"]', num_ports=256
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    host = utils_esxi.get_host(hostname, service_instance)

    spec = vim.host.VirtualSwitch.Specification()
    spec.numPorts = num_ports
    spec.mtu = mtu
    if nics:
        spec.bridge = vim.host.VirtualSwitch.BondBridge(nicDevice=nics)

    host.configManager.networkSystem.AddVirtualSwitch(vswitchName=switch_name, spec=spec)

    return {"added": switch_name}


def remove(switch_name, hostname, service_instance=None, profile=None):
    """
    Removes a vswitch on the host.

    switch_name
        The name of the vswitch to be removed.

    hostname
        The hostname where the switch exists.

    service_instance
        Use this vCenter service connection instance instead of creating a new one (optional).

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vswitch.remove switch_name=vSwitch0 host_name=host1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    host = utils_esxi.get_host(hostname, service_instance)
    host.configManager.networkSystem.RemoveVirtualSwitch(switch_name)
    return {"removed": switch_name}



# Copyright 2021 VMware, Inc.
# SPDX-License: Apache-2.0
import json
import logging

import salt.exceptions
import saltext.vmware.utils.connect as connect

log = logging.getLogger(__name__)

try:
    from pyVmomi import pbm, VmomiSupport, SoapStubAdapter, vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "storage_policy"


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def _connect_spbm_(stub, context):
    hostname = stub.host.split(":")[0]
    sessionCookie = stub.cookie.split('"')[1]
    VmomiSupport.GetRequestContext()["vcSessionCookie"] = sessionCookie

    pbmStub = SoapStubAdapter(
        host=hostname,
        path="/pbm/sdk",
        version="pbm.version.version2",
        sslContext=context,
    )
    pbmStub.cookie = stub.cookie
    pbmSi = pbm.ServiceInstance("ServiceInstance", pbmStub)
    return pbmSi


def _get_vsan_storage_policies_(pbmSi):
    """
    Example of VMOMI Object:

    (pbm.profile.CapabilityBasedProfile) {
        dynamicType = <unset>,
        dynamicProperty = (vmodl.DynamicProperty) [],
        profileId = (pbm.profile.ProfileId) {
            dynamicType = <unset>,
            dynamicProperty = (vmodl.DynamicProperty) [],
            uniqueId = '9983b8d1-d959-4bca-aa1d-364bb3c709fa'
        },
        name = 'Performance',
        description = '',
        creationTime = 2022-09-16T17:44:20.409Z,
        createdBy = 'Temporary user handle',
        lastUpdatedTime = 2022-09-16T17:44:20.409Z,
        lastUpdatedBy = 'Temporary user handle',
        profileCategory = 'REQUIREMENT',
        resourceType = (pbm.profile.ResourceType) {
            dynamicType = <unset>,
            dynamicProperty = (vmodl.DynamicProperty) [],
            resourceType = 'STORAGE'
        },
        constraints = (pbm.profile.SubProfileCapabilityConstraints) {
            dynamicType = <unset>,
            dynamicProperty = (vmodl.DynamicProperty) [],
            subProfiles = (pbm.profile.SubProfileCapabilityConstraints.SubProfile) [
                (pbm.profile.SubProfileCapabilityConstraints.SubProfile) {
                    dynamicType = <unset>,
                    dynamicProperty = (vmodl.DynamicProperty) [],
                    name = 'VSAN',
                    capability = (pbm.capability.CapabilityInstance) [
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'hostFailuresToTolerate'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'hostFailuresToTolerate',
                                    operator = <unset>,
                                    value = 1
                                }
                                ]
                            }
                        ]
                    },
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'replicaPreference'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'replicaPreference',
                                    operator = <unset>,
                                    value = 'RAID-1 (Mirroring) - Performance'
                                }
                                ]
                            }
                        ]
                    },
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'checksumDisabled'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'checksumDisabled',
                                    operator = <unset>,
                                    value = false
                                }
                                ]
                            }
                        ]
                    },
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'stripeWidth'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'stripeWidth',
                                    operator = <unset>,
                                    value = 1
                                }
                                ]
                            }
                        ]
                    },
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'forceProvisioning'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'forceProvisioning',
                                    operator = <unset>,
                                    value = false
                                }
                                ]
                            }
                        ]
                    },
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'iopsLimit'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'iopsLimit',
                                    operator = <unset>,
                                    value = 0
                                }
                                ]
                            }
                        ]
                    },
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'cacheReservation'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'cacheReservation',
                                    operator = <unset>,
                                    value = 0
                                }
                                ]
                            }
                        ]
                    },
                    (pbm.capability.CapabilityInstance) {
                        dynamicType = <unset>,
                        dynamicProperty = (vmodl.DynamicProperty) [],
                        id = (pbm.capability.CapabilityMetadata.UniqueId) {
                            dynamicType = <unset>,
                            dynamicProperty = (vmodl.DynamicProperty) [],
                            namespace = 'VSAN',
                            id = 'proportionalCapacity'
                        },
                        constraint = (pbm.capability.ConstraintInstance) [
                            (pbm.capability.ConstraintInstance) {
                                dynamicType = <unset>,
                                dynamicProperty = (vmodl.DynamicProperty) [],
                                propertyInstance = (pbm.capability.PropertyInstance) [
                                (pbm.capability.PropertyInstance) {
                                    dynamicType = <unset>,
                                    dynamicProperty = (vmodl.DynamicProperty) [],
                                    id = 'proportionalCapacity',
                                    operator = <unset>,
                                    value = 0
                                }
                                ]
                            }
                        ]
                    }
                    ],
                    forceProvision = <unset>
                }
            ]
        },
        generationId = 0,
        isDefault = false,
        systemCreatedProfileType = <unset>,
        lineOfService = <unset>
        }

    Args:
        pbmSi: PBM Service Instance

    Returns:
        list: of profile VMOMI objects
    """
    resourceType = pbm.profile.ResourceType(resourceType=pbm.profile.ResourceTypeEnum.STORAGE)

    profileManager = pbmSi.RetrieveContent().profileManager
    profileIds = profileManager.PbmQueryProfile(resourceType)
    profiles = profileManager.PbmRetrieveContent(profileIds)
    return profiles


def find(policy_name=None, service_instance=None, profile=None):
    """
    Gets storage policies. Returns list.

    policy_name
        Filter by policy name, if None returns all policies

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    pbmSi = _connect_spbm_(service_instance._stub, service_instance._stub.schemeArgs["context"])
    policies = _get_vsan_storage_policies_(pbmSi)

    result_policies = []
    for policy in policies:
        if isinstance(policy, pbm.profile.CapabilityBasedProfile):
            if policy_name is None or policy.name == policy_name:
                result_policies.append(policy)

    # make JSON representation of current policies
    # old_configs holds only the rules that are in the scope of interest (provided in argument config_input)
    result = []
    for policy in result_policies:
        policy_json = {"policyName": policy.name}
        policy_json["constraints"] = []
        try:
            for sub_profile in policy.constraints.subProfiles:
                policy_constraint = {"name": sub_profile.name}
                policy_json["constraints"].append(policy_constraint)

                for capability in sub_profile.capability:
                    for constraint in capability.constraint:
                        for prop in constraint.propertyInstance:
                            policy_constraint[prop.id] = prop.value
        except AttributeError as err:
            pass  # skip if there is no subProfiles in policy.constraints

        result.append(policy_json)

    return result


def save(policy_config, service_instance=None, profile=None):
    """
    Update policy with given configuration.

    policy_config
        Policy name and configuration values.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    pbmSi = _connect_spbm_(service_instance._stub, service_instance._stub.schemeArgs["context"])
    policies = _get_vsan_storage_policies_(pbmSi)

    policy_name = policy_config["policyName"]

    policy = None
    for profile in policies:
        if isinstance(profile, pbm.profile.CapabilityBasedProfile):
            if profile.name == policy_name:
                policy = profile
                break

    profileManager = pbmSi.RetrieveContent().profileManager

    constraints = {}
    for subProfileProps in policy_config["constraints"]:
        subProfileName = subProfileProps["name"]
        constraints[subProfileName] = {}
        for con_name, con_value in subProfileProps.items():
            if con_name != "name":
                constraints[subProfileName][con_name] = con_value

    if policy is None:
        if not policy_name:
            raise salt.exceptions.CommandExecutionError(f"Policy name is required!")

        _create_profile(profileManager, policy_name, constraints)
        return {"status": "created"}
    else:
        _update_profile(profileManager, policy.profileId, constraints)
        return {"status": "updated"}


# Create required SPBM Capability object from python dict
def _dict_to_capability(d, namespace):
    return [
        pbm.capability.CapabilityInstance(
            id=pbm.capability.CapabilityMetadata.UniqueId(namespace=namespace, id=k),
            constraint=[
                pbm.capability.ConstraintInstance(
                    propertyInstance=[pbm.capability.PropertyInstance(id=k, value=v)]
                )
            ],
        )
        for k, v in d.items()
    ]


# Update existing VM Storage Policy
def _update_profile(pm, profile_id, constraints):
    pm.PbmUpdate(
        profileId=profile_id,
        updateSpec=pbm.profile.CapabilityBasedProfileUpdateSpec(
            description=None,
            constraints=pbm.profile.SubProfileCapabilityConstraints(
                subProfiles=[
                    pbm.profile.SubProfileCapabilityConstraints.SubProfile(
                        name=name, capability=_dict_to_capability(rules, name)
                    )
                    for name, rules in constraints.items()
                ]
            ),
        ),
    )


# Create new VM Storage Policy
def _create_profile(pm, profile_name, constraints):
    pm.PbmCreate(
        createSpec=pbm.profile.CapabilityBasedProfileCreateSpec(
            name=profile_name,
            resourceType=pbm.profile.ResourceType(resourceType="STORAGE"),
            constraints=pbm.profile.SubProfileCapabilityConstraints(
                subProfiles=[
                    pbm.profile.SubProfileCapabilityConstraints.SubProfile(
                        name=name, capability=_dict_to_capability(rules, name)
                    )
                    for name, rules in constraints.items()
                ]
            ),
        )
    )



# SPDX-License-Identifier: Apache-2.0
import json
import logging

import salt.exceptions
import salt.utils.platform
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect
import saltext.vmware.utils.datastore as utils_datastore
import saltext.vmware.utils.vm as utils_vm

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim, VmomiSupport

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False

__virtualname__ = "vmware_vm"
__proxyenabled__ = ["vmware_vm"]
__func_alias__ = {"list_": "list"}


def __virtual__():
    return __virtualname__


def list_(
    service_instance=None, datacenter_name=None, cluster_name=None, host_name=None, profile=None
):
    """
    Returns virtual machines.

    datacenter_name
        Filter by this datacenter name (required when cluster is specified)

    cluster_name
        Filter by this cluster name (optional)

    host_name
        Filter by this host name (optional)

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.list
    """
    log.debug("Running vmware_vm.list")
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    return utils_vm.list_vms(
        service_instance=service_instance,
        host_name=host_name,
        cluster_name=cluster_name,
        datacenter_name=datacenter_name,
    )


def get(
    name,
    service_instance=None,
    datacenter_name=None,
    vm_properties=None,
    parent_ref=None,
    traversal_spec=None,
    profile=None,
):
    """
    Get virtual machine properties based on the traversal specs and properties list,
    returns Virtual Machine object with properties.

    service_instance
        Service instance object to access vCenter

    name
        Name of the virtual machine.

    datacenter
        Datacenter name

    vm_properties
        List of vm properties.

    traversal_spec
        Traversal Spec object(s) for searching.

    parent_ref
        Container Reference object for searching under a given object.

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.get vm-01
    """
    if service_instance is None:
        service_instance = connect.get_service_instance(config=__opts__, profile=profile)

    virtual_machine = utils_vm.get_vm_by_property(
        service_instance,
        name,
        datacenter=datacenter_name,
        vm_properties=vm_properties,
        traversal_spec=traversal_spec,
        parent_ref=parent_ref,
    )

    virtual_machine = json.loads(json.dumps(virtual_machine, cls=VmomiSupport.VmomiJSONEncoder))

    return virtual_machine


def list_templates(service_instance=None, profile=None):
    """
    Returns virtual machines tempates.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.list_templates
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    return utils_vm.list_vm_templates(service_instance)


def path(vm_name, service_instance=None, profile=None):
    """
    Returns specified virtual machine path.

    vm_name
        The name of the virtual machine.

    service_instance
        The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.path vm_name=vm01
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    vm_ref = utils_common.get_mor_by_property(
        service_instance,
        vim.VirtualMachine,
        vm_name,
    )
    return utils_common.get_path(vm_ref, service_instance)


def _deploy_ovf(name, host_name, ovf, service_instance=None, profile=None):
    """
    Helper fuctions that takes in a OVF file to create a virtual machine.

    Returns virtual machine reference.

    name
        The name of the virtual machine to be created.

    host_name
        The name of the esxi host to create the vitual machine on.

    ovf_path
        The path to the Open Virtualization Format that contains a configuration of a virtual machine.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    vms = list_(service_instance)
    if name in vms:
        raise salt.exceptions.CommandExecutionError("Duplicate virtual machine name.")

    content = service_instance.content
    manager = content.ovfManager
    spec_params = vim.OvfManager.CreateImportSpecParams(entityName=name)

    resources = utils_common.deployment_resources(host_name, service_instance)

    import_spec = manager.CreateImportSpec(
        ovf, resources["resource_pool"], resources["destination_host"].datastore[0], spec_params
    )
    errors = [e.msg for e in import_spec.error]
    if errors:
        log.exception(errors)
        raise salt.exceptions.VMwareApiError(errors)
    vm_ref = utils_vm.create_vm(
        name,
        import_spec.importSpec.configSpec,
        resources["datacenter"].vmFolder,
        resources["resource_pool"],
        resources["destination_host"],
    )
    return vm_ref


def deploy_ovf(vm_name, host_name, ovf_path, service_instance=None):
    """
    Deploy a virtual machine from an OVF

    vm_name
        The name of the virtual machine to be created.

    host_name
        The name of the esxi host to create the vitual machine on.

    ovf_path
        The path to the Open Virtualization Format that contains a configuration of a virtual machine.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.deploy_ovf vm_name=vm01 host_name=host1 ovf_path=/tmp/appliance.ovf
    """
    ovf = utils_vm.read_ovf_file(ovf_path)
    _deploy_ovf(vm_name, host_name, ovf, service_instance)
    return {"deployed": True}


def deploy_ova(vm_name, host_name, ova_path, service_instance=None):
    """
    Deploy a virtual machine from an OVA

    vm_name
        The name of the virtual machine to be created.

    host_name
        The name of the esxi host to create the vitual machine on.

    ova_path
        The path to the Open Virtualization Appliance that contains a compressed configuration of a virtual machine.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.deploy_ova vm_name=vm01 host_name=host1 ova_path=/tmp/appliance.ova
    """
    ovf = utils_vm.read_ovf_from_ova(ova_path)
    _deploy_ovf(vm_name, host_name, ovf, service_instance)
    return {"deployed": True}


def deploy_template(vm_name, template_name, host_name, service_instance=None, profile=None):
    """
    Deploy a virtual machine from a template virtual machine.

    vm_name
        The name of the virtual machine to be created.

    template_name
        The name of the template to clone from.

    host_name
        The name of the esxi host to create the vitual machine on.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.deploy_template vm_name=vm01 template_name=template1 host_name=host1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    vms = list_(service_instance)
    if vm_name in vms:
        raise salt.exceptions.CommandExecutionError("Duplicate virtual machine name.")

    template_vms = list_templates(service_instance)
    if template_name not in template_vms:
        raise salt.exceptions.CommandExecutionError("Template does not exist.")

    template = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, template_name)
    resources = utils_common.deployment_resources(host_name, service_instance)

    relospec = vim.vm.RelocateSpec()
    relospec.pool = resources["resource_pool"]

    clonespec = vim.vm.CloneSpec()
    clonespec.location = relospec

    utils_vm.clone_vm(vm_name, resources["datacenter"].vmFolder, template, clonespec)
    return {"deployed": True}


def info(vm_name=None, service_instance=None, profile=None):
    """
    Return basic info about a vSphere VM guest

    vm_name
        (optional) The name of the virtual machine to get info on.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.info vm_name=vm01
    """
    vms = []
    info = {}
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    if vm_name:
        vms.append(
            utils_common.get_mor_by_property(
                service_instance,
                vim.VirtualMachine,
                vm_name,
            )
        )

    else:
        for dc in service_instance.content.rootFolder.childEntity:
            for i in dc.vmFolder.childEntity:
                if isinstance(i, vim.VirtualMachine):
                    vms.append(i)

    for vm in vms:
        datacenter_ref = utils_common.get_parent_type(vm, vim.Datacenter)
        if vm:
            mac_address = utils_vm.get_mac_address(vm)
            network = utils_vm.get_network(vm)
            disk_bytes = utils_vm.get_disk_size(vm)
            tags = []
            for tag in vm.tag:
                try:
                    tags.append(tag.name)
                except AttributeError:
                    pass
            folder_path = utils_common.get_path(vm, service_instance)
            info[vm.summary.config.name] = {
                "guest_name": vm.summary.config.name,
                "guest_fullname": vm.summary.guest.guestFullName,
                "power_state": vm.summary.runtime.powerState,
                "ip_address": vm.summary.guest.ipAddress,
                "mac_address": mac_address,
                "uuid": vm.summary.config.uuid,
                "vm_network": network,
                "esxi_hostname": vm.summary.runtime.host.name,
                "datacenter": datacenter_ref.name,
                "cluster": vm.summary.runtime.host.parent.name,
                "tags": tags,
                "folder": folder_path,
                "moid": vm._moId,
                "num_cpu": vm.config.hardware.numCPU,
                "memory_mb": vm.config.hardware.memoryMB,
                "disk_bytes": disk_bytes,
            }
    return info


def power_state(vm_name, state, datacenter_name=None, service_instance=None, profile=None):
    """
    Manages the power state of a virtual machine.

    vm_name
        The name of the virtual machine.

    state
        The state you want the specified virtual machine in (powered-on,powered-off,suspend,reset).

    datacenter_name
        (optional) The name of the datacenter containing the virtual machine you want to manage.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.power_state vm_name=vm01 state=powered-on datacenter_name=dc1
    """
    log.trace(f"Managing power state of virtual machine {vm_name} to {state}")
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    if datacenter_name:
        dc_ref = utils_common.get_mor_by_property(service_instance, vim.Datacenter, datacenter_name)
        vm_ref = utils_common.get_mor_by_property(
            service_instance, vim.VirtualMachine, vm_name, "name", dc_ref
        )
    else:
        vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)
    if state == "powered-on" and vm_ref.summary.runtime.powerState == "poweredOn":
        result = {
            "comment": "Virtual machine is already powered on",
            "changes": {"state": vm_ref.summary.runtime.powerState},
        }
        return result
    elif state == "powered-off" and vm_ref.summary.runtime.powerState == "poweredOff":
        result = {
            "comment": "Virtual machine is already powered off",
            "changes": {"state": vm_ref.summary.runtime.powerState},
        }
        return result
    elif state == "suspend" and vm_ref.summary.runtime.powerState == "suspended":
        result = {
            "comment": "Virtual machine is already suspended",
            "changes": {"state": vm_ref.summary.runtime.powerState},
        }
        return result
    result_ref_vm = utils_vm.power_cycle_vm(vm_ref, state)
    result = {
        "comment": f"Virtual machine {state} action succeeded",
        "changes": {"state": result_ref_vm.summary.runtime.powerState},
    }
    return result


def boot_manager(
    vm_name,
    order=["cdrom", "disk", "ethernet", "floppy"],
    delay=0,
    enter_bios_setup=False,
    retry_delay=0,
    efi_secure_boot_enabled=False,
    service_instance=None,
    profile=None,
):
    """
    Manage boot option for a virtual machine

    vm_name
        The name of the virtual machine.

    order
        (List of strings) Boot order of devices. Acceptable strings: cdrom, disk, ethernet, floppy

    delay
        (integer, optional) Boot delay. When powering on or resetting, delay boot order by given milliseconds. Defaults to 0.

    enter_bios_setup
        (boolean, optional) During the next boot, force entry into the BIOS setup screen. Defaults to False.

    retry_delay
        (integer, optional) If the VM fails to find boot device, automatically retry after given milliseconds. Defaults to 0 (do not retry).

    efi_secure_boot_enabled
        (boolean, optional) Defaults to False.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.boot_manager vm_name=vm01 order='["cdrom", "disk", "ethernet"]' delay=5000 enter_bios_setup=False retry_delay=5000 efi_secure_boot_enabled=False
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    vm = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)

    boot_order_list = utils_vm.options_order_list(vm, order)

    # we removed the ability to individually set bootRetryEnabled, easily implemented if asked for
    input_opts = {
        "bootOrder": boot_order_list,
        "bootDelay": delay,
        "enterBIOSSetup": enter_bios_setup,
        "bootRetryEnabled": bool(retry_delay),
        "bootRetryDelay": retry_delay,
        "efiSecureBootEnabled": efi_secure_boot_enabled,
    }

    if utils_vm.compare_boot_options(input_opts, vm.config.bootOptions):
        return {"status": "already configured this way"}
    ret = utils_vm.change_boot_options(vm, input_opts)

    return ret


def create_snapshot(
    vm_name,
    snapshot_name,
    description="",
    include_memory=False,
    quiesce=False,
    datacenter_name=None,
    service_instance=None,
    profile=None,
):
    """
    Create snapshot of given vm.

    vm_name
        The name of the virtual machine.

    snapshot_name
        The name for the snapshot being created. Not unique

    description
        Description for the snapshot.

    include_memory
        (boolean, optional) If TRUE, a dump of the internal state of the virtual machine (basically a memory dump) is included in the snapshot.

    quiesce
        (boolean, optional) If TRUE and the virtual machine is powered on when the snapshot is taken, VMware Tools is used to quiesce the file system in the virtual machine.

    datacenter_name
        (optional) The name of the datacenter containing the virtual machine.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.create_snapshot vm_name=vm01 snapshot_name=backup_snapshot_1 description="This snapshot is a backup of vm01" include_memory=False quiesce=True datacenter_name=dc1
    """

    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    if datacenter_name:
        dc_ref = utils_common.get_mor_by_property(service_instance, vim.Datacenter, datacenter_name)
        vm_ref = utils_common.get_mor_by_property(
            service_instance, vim.VirtualMachine, vm_name, "name", dc_ref
        )
    else:
        vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)

    snapshot = utils_vm.create_snapshot(vm_ref, snapshot_name, description, include_memory, quiesce)

    if isinstance(snapshot, vim.vm.Snapshot):
        return {"snapshot": "created"}
    else:
        return {"snapshot": "failed to create"}


def destroy_snapshot(
    vm_name,
    snapshot_name,
    snapshot_id=None,
    remove_children=False,
    datacenter_name=None,
    service_instance=None,
    profile=None,
):
    """
    Destroy snapshot of given vm.

    vm_name
        The name of the virtual machine.

    snapshot_name
        The name for the snapshot being destroyed. Not unique

    snapshot_id
        (optional) ID of snapshot to be destroyed.

    remove_children
        (optional, Bool) Remove snapshots below snapshot being removed in tree.

    datacenter_name
        (optional) The name of the datacenter containing the virtual machine.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.destroy_snapshot vm_name=vm01 snapshot_name=backup_snapshot_1 snapshot_id=1 remove_children=False datacenter_name=dc1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    if datacenter_name:
        dc_ref = utils_common.get_mor_by_property(service_instance, vim.Datacenter, datacenter_name)
        vm_ref = utils_common.get_mor_by_property(
            service_instance, vim.VirtualMachine, vm_name, "name", dc_ref
        )
    else:
        vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)

    snap_ref = utils_vm.get_snapshot(vm_ref, snapshot_name, snapshot_id)
    utils_vm.destroy_snapshot(snap_ref.snapshot, remove_children)
    return {"snapshot": "destroyed"}


def snapshot(vm_name, datacenter_name=None, service_instance=None, profile=None):
    """
    Return info about a virtual machine snapshots

    vm_name
        (optional) The name of the virtual machine to get info on.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.snapshot vm_name=vm01 datacenter_name=dc1
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )

    if datacenter_name:
        dc_ref = utils_common.get_mor_by_property(service_instance, vim.Datacenter, datacenter_name)
        vm_ref = utils_common.get_mor_by_property(
            service_instance, vim.VirtualMachine, vm_name, "name", dc_ref
        )
    else:
        vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)

    snapshots = utils_vm.get_snapshots(vm_ref)

    return {"snapshots": snapshots}


def relocate(
    vm_name,
    new_host_name,
    datastore_name,
    datacenter_name=None,
    service_instance=None,
    profile=None,
):
    """
    Relocates a virtual machine to the location specified.

    vm_name
        The name of the virtual machine to relocate.

    new_host_name
        The name of the host you want to move the virtual machine to.

    datastore_name
        The name of the datastore you want to move the virtual machine to.

    datacenter_name
        The name of the datacenter containing the datastore.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.relocate vm_name=vm01 new_host_name=host1 datastore_name=ds01
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)
    resources = utils_common.deployment_resources(new_host_name, service_instance)
    assert isinstance(datastore_name, str)
    datastores = utils_datastore.get_datastores(
        service_instance, datastore_name=datastore_name, datacenter_name=datacenter_name
    )
    datastore_ref = datastores[0] if datastores else None
    ret = utils_vm.relocate(
        vm_ref, resources["destination_host"], datastore_ref, resources["resource_pool"]
    )
    if ret == "success":
        return {"virtual_machine": "moved"}
    return {"virtual_machine": "failed to move"}


def network(vm_name, service_instance=None, profile=None):
    """
    Retreives the networking for a virtual machine.

    vm_name
        The name of the virtual machine to relocate.

    service_instance
        (optional) The Service Instance from which to obtain managed object references.

    profile
        Profile to use (optional)

    CLI Example:

    .. code-block:: bash

        salt '*' vmware_vm.network vm_name=vm01
    """
    ret = {}

    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)

    network_refs = vm_ref.network

    for network in network_refs:
        ret[network.name] = {}

        try:
            ret[network.name]["config"] = network.config
        except AttributeError:
            # Fetch the port groups via host configuration attached to the virtual machine
            host = vm_ref.summary.runtime.host
            if host.config:
                for portgroup in host.config.network.portgroup:
                    if network.name == portgroup.spec.name:
                        ret[network.name]["config"] = portgroup.spec

    ret = json.loads(json.dumps(ret, cls=VmomiSupport.VmomiJSONEncoder))

    return ret


def get_mks_ticket(vm_name, ticket_type, service_instance=None, profile=None):
    """
    Get ticket of virtual machine of passed object type.

    vm_name
        The name of the virtual machine which has tickets. VM names can be
        found in ``vmware_vm.list``.

    ticket_type
        Type of ticket - device, guestControl, guestIntegrity, mks, or webmks.

        See https://vdc-download.vmware.com/vmwb-repository/dcr-public/3325c370-b58c-4799-99ff-58ae3baac1bd/45789cc5-aba1-48bc-a320-5e35142b50af/doc/vim.VirtualMachine.TicketType.html

        salt '*' vmware_vm.get_mks_ticket vm_name=vm01 ticket_type=webmks
    """
    if service_instance is None:
        service_instance = connect.get_service_instance(config=__opts__, profile=profile)

    log.info(f"Acquiring ticket {ticket_type} for {vm_name}")
    vm_ref = utils_common.get_mor_by_property(service_instance, vim.VirtualMachine, vm_name)
    if vm_ref:
        ticket = vm_ref.AcquireTicket(ticket_type)
        return json.loads(json.dumps(ticket, cls=VmomiSupport.VmomiJSONEncoder))
    return {}



# SPDX-License-Identifier: Apache-2.0
"""
Define the required entry-points functions in order for Salt to know
what and from where it should load this extension's loaders
"""
from . import PACKAGE_ROOT



# SPDX-License-Identifier: Apache-2.0
import logging

import saltext.vmware.modules.folder as m_folder
import saltext.vmware.utils.common as utils_common
import saltext.vmware.utils.connect as connect

log = logging.getLogger(__name__)

try:
    from pyVmomi import vim

    HAS_PYVMOMI = True
except ImportError:
    HAS_PYVMOMI = False


__virtualname__ = "vmware_folder"
__proxyenabled__ = ["vmware_folder"]


def __virtual__():
    if not HAS_PYVMOMI:
        return False, "Unable to import pyVmomi module."
    return __virtualname__


def manage(name, task, dc_name, type, service_instance=None, profile=None):
    """
    Creates or destroy a folder on a given datacenter.

    name
        Name of folder.

    task
        create or destroy

    dc_name
        Name of datacenter where folder will be created.

    type
        (vm, host, datastore, network) Type of folder to be created.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}
    task_string = "created" if (task == "create") else "destroyed"

    if __opts__["test"]:
        ret["changes"] = {"new": f"folder {name} will be {task_string}"}
        ret["comment"] = "These options are set to change."
        return ret

    if task == "create":
        result = m_folder.create(name, dc_name, type, service_instance)
        ret["changes"] = {"new": f"folder {name} {task_string}"}
    elif task == "destroy":
        result = m_folder.destroy(name, dc_name, type, service_instance)
        ret["changes"] = {"new": f"folder {name} {task_string}"}
    else:
        ret["comment"] = "invalid task"
        ret["result"] = False
        return ret

    ret["comment"] = result["status"]
    return ret


def rename(name, new_folder_name, dc_name, type, service_instance=None, profile=None):
    """
    Rename a folder on a given datacenter.

    name
        Name of folder.

    new_folder_name
        Name to rename folder.

    dc_name
        Name of datacenter where folder will be renamed.

    type
        (vm, host, datastore, network) Type of folder to be renamed.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if __opts__["test"]:
        ret["changes"] = {"new": f"folder {name} will be renamed {new_folder_name}"}
        ret["comment"] = "These options are set to change."
        return ret

    result = m_folder.rename(name, new_folder_name, dc_name, type, service_instance)
    ret["changes"] = {"new": f"folder {name} renamed {new_folder_name}"}
    ret["comment"] = result["status"]
    return ret


def move(name, destination_folder_name, dc_name, type, service_instance=None, profile=None):
    """
    Move a folder on a given datacenter.

    folder_name
        Name of folder.

    destination_folder_name
        Destination folder for named folder.

    dc_name
        Name of datacenter where folder will be moved.

    type
        (vm, host, datastore, network) Type of folder to be moved.

    service_instance
        Use this vCenter service connection instance instead of creating a new one. (optional).

    profile
        Profile to use (optional)
    """
    service_instance = service_instance or connect.get_service_instance(
        config=__opts__, profile=profile
    )
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if __opts__["test"]:
        ret["changes"] = {"new": f"folder {name} will be moved to {destination_folder_name}"}
        ret["comment"] = "These options are set to change."
        return ret

    result = m_folder.move(name, destination_folder_name, dc_name, type, service_instance)
    ret["changes"] = {"new": f"folder {name} moved to {destination_folder_name}"}
    ret["comment"] = result["status"]
    return ret
